from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum, auto
from typing import TYPE_CHECKING

from .intrinsics import Intrinsic

if TYPE_CHECKING:
    from gofra.lexer import Token
    from gofra.typecheck.types import GofraType

type OperatorOperand = int | str | None | Intrinsic


class OperatorType(IntEnum):
    PUSH_INTEGER = auto()
    PUSH_STRING = auto()
    PUSH_MEMORY_POINTER = auto()  # Static segment memory pointer

    INTRINSIC = auto()

    IF = auto()
    WHILE = auto()
    DO = auto()
    END = auto()

    FUNCTION_RETURN = auto()
    FUNCTION_CALL = auto()


@dataclass(frozen=False)
class Operator:
    type: OperatorType
    token: Token
    operand: OperatorOperand

    jumps_to_operator_idx: int | None = field(default=None)

    syscall_optimization_omit_result: bool = field(default=False)
    syscall_optimization_injected_args: list[int | None] | None = None

    has_optimizations: bool = field(default=False)
    infer_type_after_optimization: GofraType | None = field(default=None)

    def __repr__(self) -> str:
        return f"OP<{self.type.name}>"

    def is_syscall(self) -> bool:
        return self.type == OperatorType.INTRINSIC and self.operand in (
            Intrinsic.SYSCALL0,
            Intrinsic.SYSCALL1,
            Intrinsic.SYSCALL2,
            Intrinsic.SYSCALL3,
            Intrinsic.SYSCALL4,
            Intrinsic.SYSCALL5,
            Intrinsic.SYSCALL6,
        )

    def get_syscall_arguments_count(self) -> int:
        assert self.is_syscall()
        assert isinstance(self.operand, Intrinsic)
        return int(self.operand.name[-1]) + 1



"""Gofra programming language.

Provides toolchain including CLI, compiler etc.
"""

from .assembler import assemble_program
from .gofra import process_input_file

__all__ = [
    "assemble_program",
    "process_input_file",
]



from collections.abc import Generator
from dataclasses import dataclass

from gofra.lexer.tokens import Token


@dataclass(frozen=True)
class Macro:
    token: Token

    name: str

    tokens: list[Token]

    def as_tokenizer(self) -> Generator[Token]:
        yield from self.tokens



from string import hexdigits

from gofra.exceptions import GofraError
from gofra.lexer.helpers import HEXADECIMAL_MARK

from .tokens import TokenLocation


class LexerError(GofraError):
    """General error within lexer.

    Should not be used directly as not provides information about error source!
    """

    def __repr__(self) -> str:
        return """General lexer error occurred. 

Please open an issue about that undocumented behavior!
"""


class LexerAmbigiousHexadecimalAlphabetError(LexerError):
    def __init__(
        self,
        *args: object,
        hexadecimal_raw: str,
        number_location: TokenLocation,
    ) -> None:
        super().__init__(*args)
        self.number_location = number_location
        self.hexadecimal_raw = hexadecimal_raw

    def __repr__(self) -> str:
        return f"""Ambigious hexadecimal alphabet inside '{self.hexadecimal_raw}' at {self.number_location}!

Hexadecimal numbers ({HEXADECIMAL_MARK}) must consinst of 16-base alphabet ({hexdigits})"""


class LexerUnclosedCharacterQuoteError(LexerError):
    def __init__(self, *args: object, open_quote_location: TokenLocation) -> None:
        super().__init__(*args)
        self.open_quote_location = open_quote_location

    def __repr__(self) -> str:
        return f"""Unclosed character quote at {self.open_quote_location}!

Did you forgot to close an character?"""


class LexerExcessiveCharacterLengthError(LexerError):
    def __init__(
        self,
        *args: object,
        excess_begins_at: TokenLocation,
        excess_by_count: int,
    ) -> None:
        super().__init__(*args)
        self.excess_begins_at = excess_begins_at
        self.excess_by_count = excess_by_count

    def __repr__(self) -> str:
        return f"""Excessive amount of symbols in character at {self.excess_begins_at}!

Expected only one symbol in character but got {self.excess_by_count}!

Did you accidentally mix up characters and strings?"""


class LexerEmptyCharacterError(LexerError):
    def __init__(self, *args: object, open_quote_location: TokenLocation) -> None:
        super().__init__(*args)
        self.open_quote_location = open_quote_location

    def __repr__(self) -> str:
        return f"""Empty character at {self.open_quote_location}!

Expected single symbol in character!

Did you forgot to enter an character?"""


class LexerUnclosedStringQuoteError(LexerError):
    def __init__(
        self,
        *args: object,
        open_quote_location: TokenLocation,
    ) -> None:
        super().__init__(*args)
        self.open_quote_location = open_quote_location

    def __repr__(self) -> str:
        return f"""Unclosed string quote at {self.open_quote_location}

Did you forgot to close string or mistyped string quote?"""



from enum import IntEnum, auto


class Keyword(IntEnum):
    """Words that are related to language due to internal implementation like loops or parsing stage."""

    IF = auto()

    WHILE = auto()
    DO = auto()

    END = auto()

    MEMORY = auto()

    EXTERN = auto()
    INLINE = auto()
    GLOBAL = auto()

    FUNCTION = auto()
    FUNCTION_RETURN = auto()
    FUNCTION_CALL = auto()

    # Preprocessor
    PP_IFDEF = auto()
    PP_ENDIF = auto()
    PP_INCLUDE = auto()
    PP_MACRO = auto()


WORD_TO_KEYWORD = {
    "if": Keyword.IF,
    "while": Keyword.WHILE,
    "do": Keyword.DO,
    "end": Keyword.END,
    "extern": Keyword.EXTERN,
    "call": Keyword.FUNCTION_CALL,
    "return": Keyword.FUNCTION_RETURN,
    "func": Keyword.FUNCTION,
    "inline": Keyword.INLINE,
    "memory": Keyword.MEMORY,
    "global": Keyword.GLOBAL,
    # Preprocessor
    "#ifdef": Keyword.PP_IFDEF,
    "#endif": Keyword.PP_ENDIF,
    "#include": Keyword.PP_INCLUDE,
    "#macro": Keyword.PP_MACRO,
}
KEYWORD_TO_NAME = {v: k for k, v in WORD_TO_KEYWORD.items()}



"""Assembly abstraction layer that hides declarative assembly OPs into functions that generates that for you."""

from __future__ import annotations

from typing import TYPE_CHECKING, assert_never

from .registers import (
    AMD64_LINUX_ABI_ARGUMENTS_REGISTERS,
    AMD64_LINUX_ABI_RETVAL_REGISTER,
    AMD64_LINUX_SYSCALL_ARGUMENTS_REGISTERS,
    AMD64_LINUX_SYSCALL_NUMBER_REGISTER,
)

if TYPE_CHECKING:
    from gofra.codegen.backends.general import CODEGEN_GOFRA_ON_STACK_OPERATIONS

    from ._context import AMD64CodegenContext
    from .registers import AMD64_GP_REGISTERS


def drop_cells_from_stack(context: AMD64CodegenContext, *, cells_count: int) -> None:
    """Drop stack cells from stack.

    Stack alignment should be review, but `popq` unlike `pop` will shift by 64 bits
    """
    assert cells_count > 0, "Tried to drop negative cells count from stack"
    for _ in range(cells_count):
        context.write("popq rax")  # Use zero regiseter?


def pop_cells_from_stack_into_registers(
    context: AMD64CodegenContext,
    *registers: AMD64_GP_REGISTERS,
) -> None:
    """Pop cells from stack and store into given registers.

    Stack alignment should be review, but probably this codegen ensure with `popq`/`pushq` that all cells are 64 bits
    """
    assert registers, "Expected registers to store popped result into!"

    for register in registers:
        context.write(f"popq {register}")


def push_register_onto_stack(
    context: AMD64CodegenContext,
    register: AMD64_GP_REGISTERS,
) -> None:
    """Store given register onto stack under current stack pointer."""
    context.write(f"pushq {register}")


def store_integer_into_register(
    context: AMD64CodegenContext,
    register: AMD64_GP_REGISTERS,
    value: int,
) -> None:
    """Store given value into given register (as QWORD)."""
    context.write(f"movq ${value}, {register}")


def push_integer_onto_stack(
    context: AMD64CodegenContext,
    value: int,
) -> None:
    """Push given integer onto stack.

    Negative numbers is dissalowed.

    TODO(@kirillzhosul, @stepanzubkov): Negative numbers IS dissalowed
    TODO(@kirillzhosul, @stepanzubkov): Review max and etc like in AARCH64_MacOS
    """
    assert value >= 0, "Tried to push negative integer onto stack!"

    store_integer_into_register(context, register="rax", value=value)
    push_register_onto_stack(context, register="rax")


def push_static_address_onto_stack(
    context: AMD64CodegenContext,
    segment: str,
) -> None:
    """Push executable static memory addresss onto stack with page dereference."""
    context.write(f"leaq {segment}(rip), rax")
    push_register_onto_stack(context, register="rax")


def initialize_static_data_section(
    context: AMD64CodegenContext,
    static_data_section: list[tuple[str, str | int]],
) -> None:
    """Initialize data section fields with given values.

    Section is an tuple (label, data)
    Data is an string (raw ASCII) or number (zeroed memory blob)
    TODO(@kirillzhosul, @stepanzubkov): Review alignment for data sections.
    """
    context.fd.write(".section .data\n")

    for name, data in static_data_section:
        if isinstance(data, str):
            context.fd.write(f'{name}: .asciz "{data}"\n')
            continue
        context.fd.write(f"{name}: .space {data}\n")


def ipc_syscall_linux(
    context: AMD64CodegenContext,
    *,
    arguments_count: int,
    store_retval_onto_stack: bool,
    injected_args: list[int | None] | None,
) -> None:
    """Call system (syscall) and apply IPC ABI convention to arguments."""
    assert not injected_args or len(injected_args) == arguments_count + 1
    registers_to_load = (
        AMD64_LINUX_SYSCALL_NUMBER_REGISTER,
        *AMD64_LINUX_SYSCALL_ARGUMENTS_REGISTERS[:arguments_count][::-1],
    )

    if not injected_args:
        injected_args = [None for _ in range(arguments_count + 1)]

    for injected_argument, register in zip(injected_args, registers_to_load):
        if injected_argument is not None:
            # Register injected and infered from stack
            store_integer_into_register(
                context,
                register=register,
                value=injected_argument,
            )
            continue
        pop_cells_from_stack_into_registers(context, register)

    context.write("syscall")

    if store_retval_onto_stack:
        # Mostly related to optimizations above if we dont want to store result
        push_register_onto_stack(context, AMD64_LINUX_ABI_RETVAL_REGISTER)


def function_end_with_epilogue(context: AMD64CodegenContext) -> None:
    """Functions epilogue at the end. Restores required fields (like stack-pointer)."""
    context.write("ret")


def function_begin_with_prologue(
    context: AMD64CodegenContext,
    *,
    function_name: str,
    as_global_linker_symbol: bool,
) -> None:
    """Begin an function symbol with prologue with preparing required (like stack-pointer).

    TODO(@kirillzhosul, @stepanzubkov): Review alignment for executable sections.
    """
    if as_global_linker_symbol:
        context.fd.write(f".global {function_name}\n")
    context.fd.write(f"{function_name}:\n")


def call_function_block(
    context: AMD64CodegenContext,
    function_name: str,
    *,
    abi_ffi_push_retval_onto_stack: bool,
    abi_ffi_arguments_count: int,
) -> None:
    """Call an function with preparing required fields (like stack-pointer).

    Also allows to call ABI/FFI function with providing arguments and return value (retval) via:
    `abi_ffi_push_retval_onto_stack` / `abi_ffi_arguments_count`
    """
    assert abi_ffi_arguments_count >= 0, "FFI arguments count cannot go negative"
    if abi_ffi_arguments_count:
        arguments = abi_ffi_arguments_count
        registers = AMD64_LINUX_ABI_ARGUMENTS_REGISTERS[:arguments][::-1]
        pop_cells_from_stack_into_registers(context, *registers)

    context.write(f"call {function_name}")

    if abi_ffi_push_retval_onto_stack:
        push_register_onto_stack(context, AMD64_LINUX_ABI_RETVAL_REGISTER)


def store_into_memory_from_stack_arguments(context: AMD64CodegenContext) -> None:
    """Store value from into memory pointer, pointer and value acquired from stack."""
    pop_cells_from_stack_into_registers(context, "rax", "rbx")
    context.write("movq rax, (rbx)")


def load_memory_from_stack_arguments(context: AMD64CodegenContext) -> None:
    """Load memory as value using arguments from stack."""
    pop_cells_from_stack_into_registers(context, "rax")
    context.write("movq (rax), rax")
    push_register_onto_stack(context, "rax")


def perform_operation_onto_stack(
    context: AMD64CodegenContext,
    operation: CODEGEN_GOFRA_ON_STACK_OPERATIONS,
) -> None:
    """Perform *math* operation onto stack (pop arguments and push back result)."""
    is_unary = operation in ("++", "--")
    registers = ("rax",) if is_unary else ("rax", "rbx")
    pop_cells_from_stack_into_registers(context, *registers)

    match operation:
        case "+":
            context.write("addq rbx, rax")
        case "-":
            context.write("subq rax, rbx")
        case "*":
            context.write("mulq rbx, rax")
        case "//":
            raise NotImplementedError
        case "%":
            raise NotImplementedError
        case "++":
            context.write("incq rax")
        case "--":
            context.write("decq rax")
        case "!=" | ">=" | "<=" | "<" | ">" | "==":
            logic_op = {
                "!=": "ne",
                ">=": "ge",
                "<=": "le",
                "<": "l",
                ">": "g",
                "==": "e",
            }

            context.write(
                "cmpq rbx, rax",
                "xorq rax, rax",
                f"set{logic_op[operation]}b al",
            )
        case _:
            assert_never()
    push_register_onto_stack(context, "rax")


def evaluate_conditional_block_on_stack_with_jump(
    context: AMD64CodegenContext,
    jump_over_label: str,
) -> None:
    """Evaluate conditional block by popping current value under SP againts zero.

    If condition is false (value on stack) then jump out that conditional block to `jump_over_label`
    """
    pop_cells_from_stack_into_registers(context, "rax")
    context.write(
        "cmpq $0, rax",
        f"jeq {jump_over_label}",
    )



"""Gofra core entry."""

from collections.abc import Iterable
from pathlib import Path

from gofra.context import ProgramContext
from gofra.lexer import tokenize_file
from gofra.parser import parse_file
from gofra.preprocessor import preprocess_file


def process_input_file(
    filepath: Path,
    include_paths: Iterable[Path],
) -> ProgramContext:
    """Core entry for Gofra API.

    Compiles given filepath down to `IR` into `ProgramContext`.
    Maybe assembled into executable/library/object/etc... via `assemble_program`

    Does not provide optimizer or type checker.
    """
    lexer = tokenize_file(filepath)
    preprocessor = preprocess_file(filepath, lexer, include_paths)
    parser_context, entry_point = parse_file(preprocessor)
    return ProgramContext.from_parser_context(parser_context, entry_point)



"""Assembler package that links and assembles generated code into final executable.

Tools used for assembly is different for specified target
"""

from .assembler import assemble_program

__all__ = ["assemble_program"]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING

from gofra.assembler.assembler import assemble_program
from gofra.cli.errors import cli_gofra_error_handler
from gofra.cli.output import CLIColor, cli_message
from gofra.exceptions import GofraError
from gofra.gofra import process_input_file

from .arguments import CLIArguments, parse_cli_arguments

if TYPE_CHECKING:
    from collections.abc import Generator


class TestStatus(Enum):
    SKIPPED = auto()

    ERROR = auto()
    SUCCESS = auto()


@dataclass(frozen=False)
class Test:
    status: TestStatus
    path: Path
    error: GofraError | None = None


COLORS = {
    TestStatus.SUCCESS: CLIColor.GREEN,
    TestStatus.ERROR: CLIColor.RED,
    TestStatus.SKIPPED: CLIColor.RESET,
}
ICONS = {
    TestStatus.SUCCESS: "+",
    TestStatus.ERROR: "-",
    TestStatus.SKIPPED: ".",
}


def cli_entry_point() -> None:
    """CLI main entry."""
    with cli_gofra_error_handler():
        args = parse_cli_arguments()
        cli_process_testkit_runner(args)


def cli_process_testkit_runner(args: CLIArguments) -> None:
    """Process full testkit toolchain."""
    cli_message(level="INFO", text="Searching test files...")

    test_paths = tuple(search_test_case_files(args.directory))
    cli_message(
        level="INFO",
        text=f"Found {len(test_paths)} test case files.",
    )

    test_matrix: list[Test] = []
    for test_path in test_paths:
        try:
            context = process_input_file(filepath=test_path, include_paths=[])
        except GofraError as e:
            test_matrix.append(Test(status=TestStatus.ERROR, path=test_path, error=e))
            continue
        assemble_program(
            context,
            Path(
                args.build_cache_dir
                / f"._testkit__build__{test_path.with_suffix('').name}",
            ),
            output_format="executable",
            target="aarch64-darwin",
            build_cache_dir=args.build_cache_dir,
            verbose=True,
            additional_assembler_flags=[],
            additional_linker_flags=[],
            delete_build_cache_after_compilation=True,
        )
        test_matrix.append(Test(status=TestStatus.SUCCESS, path=test_path))
    display_test_matrix(test_matrix)
    display_test_errors(test_matrix)


def search_test_case_files(directory: Path) -> Generator[Path]:
    return directory.glob("test_*.gof", case_sensitive=False)


def display_test_matrix(matrix: list[Test]) -> None:
    for test in matrix:
        color = COLORS[test.status]
        icon = ICONS[test.status]
        print(f"{color}{icon}", test.path)


def display_test_errors(matrix: list[Test]) -> None:
    if any(test.error for test in matrix):
        cli_message("ERROR", "While running tests, some errors were occured:")
    for test in matrix:
        if test.error is None:
            continue
        cli_message("INFO", f"While testing `{test.path}`:")
        cli_message("ERROR", repr(test.error))



"""Consts and types related to AARCH64 registers and architecture (including FFI/ABI/IPC)."""

from __future__ import annotations

from typing import Literal

####
# Bare AARCH64 related
####

# Stack is aligment to specified bytes count
# Each cell of an stack must be within that size
# Pushing >2 cells onto stack will lead to cell overflow due to language stack nature.
AARCH64_STACK_ALIGNMENT = 16
AARCH64_STACK_ALINMENT_BIN = 4  # 2 ** 4 -> AARCH64_STACK_ALIGNMENT

# Bits count (size) for different word types
AARCH64_HALF_WORD_BITS = 0xFFFF  # 4 bytes (16 bits)
AARCH64_DOUBLE_WORD_BITS = 0xFFFF_FFFF_FFFF_FFFF  # 16 bytes (64 bits)

# Registers specification for AARCH64
# Skips some of registers (X8-X15, X18-X30) due to currently being unused
type AARCH64_ABI_X_REGISTERS = Literal["X0", "X1", "X2", "X3", "X4", "X5", "X6", "X7"]
type AARCH64_ABI_W_REGISTERS = Literal["W0", "W1", "W2", "W3", "W4", "W5", "W6", "W7"]
type AARCH64_ABI_REGISTERS = AARCH64_ABI_X_REGISTERS | AARCH64_ABI_W_REGISTERS
type AARCH64_IPC_REGISTERS = Literal["X16", "X17"]
type AARCH64_GP_REGISTERS = AARCH64_ABI_REGISTERS | AARCH64_IPC_REGISTERS


####
# MacOS related
####

# MacOS syscall convention
AARCH64_MACOS_SYSCALL_NUMBER_REGISTER: AARCH64_IPC_REGISTERS = "X16"
AARCH64_MACOS_ABI_RETVAL_REGISTER: AARCH64_ABI_REGISTERS = "X0"
AARCH64_MACOS_ABI_ARGUMENT_REGISTERS: tuple[AARCH64_ABI_REGISTERS, ...] = (
    "X0",
    "X1",
    "X2",
    "X3",
    "X4",
    "X5",
    "X6",
    "X7",
)

# Epilogue
AARCH64_MACOS_EPILOGUE_EXIT_CODE = 0
AARCH64_MACOS_EPILOGUE_EXIT_SYSCALL_NUMBER = 1



from gofra.codegen.targets import TARGET_T

from .backends import (
    CodeGeneratorBackend,
    generate_aarch64_macos_backend,
    generate_amd64_linux_backend,
)
from .exceptions import CodegenUnsupportedBackendTargetPairError


def get_backend_for_target(
    target: TARGET_T,
) -> CodeGeneratorBackend:
    """Get code generator backend for specified ARCHxOS pair."""
    match target:
        case "aarch64-darwin":
            return generate_aarch64_macos_backend
        case "x86_64-linux":
            return generate_amd64_linux_backend
        case _:
            raise CodegenUnsupportedBackendTargetPairError(target=target)



from pathlib import Path

from gofra.exceptions import GofraError


class IOFileDoesNotExistsError(GofraError):
    def __init__(self, *args: object, path: Path) -> None:
        super().__init__(*args)
        self.path = path

    def __repr__(self) -> str:
        return f"""File not found (I/O error)

Tried to open file at path: `{self.path}`
(Resolves to: `{self.path.resolve()})

Please ensure that requested file/symlink exists!"""


class IOFileNotAnFileError(GofraError):
    def __init__(self, *args: object, path: Path) -> None:
        super().__init__(*args)
        self.path = path

    def __repr__(self) -> str:
        return f"""File is not an file 
(Is an directory?: {"Yes" if self.path.is_dir() else "No"})

Tried to open file at path: `{self.path}`
(Resolves to: `{self.path.resolve()})

Please ensure that requested file is an file and not an directory or anything else!"""



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.lexer.keywords import Keyword
from gofra.lexer.tokens import Token, TokenType

from ._state import PreprocessorState
from .conditions import resolve_conditional_block_from_token
from .include import resolve_include_from_token_into_state
from .macros.preprocessor import (
    define_macro_block_from_token,
    try_resolve_macro_reference_from_token,
)

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from pathlib import Path


def preprocess_file(
    path: Path,
    lexer: Generator[Token],
    include_search_paths: Iterable[Path],
) -> Generator[Token]:
    """Preprocess given lexer token stream by resolving includes, CTE/macros.

    Simply, wraps an lexer into another `lexer` and preprocess on the fly.
    """
    state = PreprocessorState(
        path=path,
        lexer=lexer,
        include_search_paths=include_search_paths,
    )

    for token in state.tokenizer:
        match token:
            case Token(type=TokenType.KEYWORD, value=Keyword.PP_INCLUDE):
                resolve_include_from_token_into_state(token, state)
            case Token(type=TokenType.KEYWORD, value=Keyword.PP_MACRO):
                define_macro_block_from_token(token, state)
            case Token(type=TokenType.WORD):
                if try_resolve_macro_reference_from_token(token, state):
                    continue
                yield token
            case Token(type=TokenType.KEYWORD, value=Keyword.PP_IFDEF):
                resolve_conditional_block_from_token(token, state)
            case Token(type=TokenType.KEYWORD, value=Keyword.PP_ENDIF):
                # We dont yield preprocessor endif as preprocessor will resolve it by itself via conditional block resolver
                pass
            case _:
                yield token



from __future__ import annotations

from string import hexdigits
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable


STRING_QUOTE = '"'
CHARACTER_QUOTE = "'"
SINGLE_LINE_COMMENT = "//"
HEXADECIMAL_MARK = "0x"
ESCAPE_SYMBOL = "\\"


def unescape_string(string: str) -> str:
    """Remove all terminations within string (escape it)."""
    return string.encode("unicode_escape").decode("UTF-8")


def is_valid_hexadecimal(text: str) -> bool:
    """Is given raw text cans be parsed as hexadecimal?."""
    return all(c in hexdigits for c in text[len(HEXADECIMAL_MARK) :])


def is_valid_integer(text: str) -> bool:
    """Is given raw text can be parsed as integer?."""
    return text.isdigit()


def find_word_start(text: str, start: int) -> int:
    """Find start column index of an word."""
    return _find_column(text, start, lambda s: not s.isspace())


def find_word_end(text: str, start: int) -> int:
    """Find end column index of an word."""
    return _find_column(text, start, str.isspace)


def find_string_end(string: str, start: int) -> int | None:
    """Find index where given string ends (close quote) or None if not closed properly."""
    idx = start
    idx_end = len(string)

    prev = string[idx]

    while idx < idx_end:
        current = string[idx]
        if current == STRING_QUOTE and prev != ESCAPE_SYMBOL:
            break

        prev = current
        idx += 1

    if idx >= idx_end:
        return None
    return idx + 1


def _find_column(text: str, start: int, predicate: Callable[[str], bool]) -> int:
    """Find index of an column by predicate. E.g `.index()` but with predicate."""
    end = len(text)
    while start < end and not predicate(text[start]):
        start += 1
    return start



from typing import IO, Protocol

from gofra.context import ProgramContext


class CodeGeneratorBackend(Protocol):
    """Base code generator backend protocol.

    All backends inherited from this protocol.
    """

    def __call__(
        self,
        fd: IO[str],
        program: ProgramContext,
    ) -> None: ...



"""Code generation backend module.

Provides code generation backends (codegens) for emitting assembly from IR.
"""

from .aarch64_macos import generate_aarch64_macos_backend
from .amd64_linux import generate_amd64_linux_backend
from .base import CodeGeneratorBackend

__all__ = [
    "CodeGeneratorBackend",
    "generate_aarch64_macos_backend",
    "generate_amd64_linux_backend",
]



from pathlib import Path

from gofra.codegen.targets import TARGET_T
from gofra.context import ProgramContext

from .get_backend import get_backend_for_target


def generate_code_for_assembler(
    output_path: Path,
    context: ProgramContext,
    target: TARGET_T,
) -> None:
    """Generate assembly from given program context and specified ARCHxOS pair into given file."""
    backend = get_backend_for_target(target)

    output_path.parent.mkdir(exist_ok=True)
    with output_path.open(
        mode="w",
        errors="strict",
        buffering=1,
        newline="",
        encoding="UTF-8",
    ) as fd:
        return backend(fd, context)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from gofra.parser.operators import Operator

from .exceptions import (
    TypecheckInvalidFunctionArgumentTypeError,
    TypecheckInvalidOperatorArgumentTypeError,
    TypecheckNotEnoughFunctionArgumentsError,
    TypecheckNotEnoughOperatorArgumentsError,
)
from .types import GofraType as T

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from gofra.parser.functions.function import Function


@dataclass(frozen=False)
class TypecheckContext:
    """Context for type checking which only required for internal usages."""

    # Typechecker is an emulated type stack, e.g 1 2 would produce [INT, INT] ino type stack
    emulated_stack_types: MutableSequence[T]

    def push_types(self, *types: T) -> None:
        """Push given types onto emulated typeS stack."""
        self.emulated_stack_types.extend(types)

    def raise_for_enough_arguments(
        self,
        operator_or_function: Operator | Function,
        inside_function: Function,
        required_args: int,
        operator: Operator | None = None,
    ) -> None:
        """Expect that stack has N arguments."""
        stack_size = len(self.emulated_stack_types)
        if stack_size < required_args:
            if isinstance(operator_or_function, Operator):
                raise TypecheckNotEnoughOperatorArgumentsError(
                    operator=operator_or_function,
                    types_on_stack=self.emulated_stack_types,
                    required_args=required_args,
                )
            assert operator, "Expected call operator"
            raise TypecheckNotEnoughFunctionArgumentsError(
                function=operator_or_function,
                callee_function=inside_function,
                types_on_stack=self.emulated_stack_types,
                called_from_operator=operator,
            )

    def pop_type_from_stack(self) -> T:
        """Pop current type on the stack."""
        return self.emulated_stack_types.pop()

    def consume_n_arguments(self, args_to_consume: int) -> None:
        """Pop N arguments from stack of any type."""
        for _ in range(args_to_consume):
            self.pop_type_from_stack()

    def raise_for_arguments(
        self,
        operator_or_function: Operator | Function,
        inside_function: Function,
        *expected_types: T,
        operator: Operator | None = None,
    ) -> None:
        """Expect given arguments and count and their types should match.

        Types are reversed so call will look like original stack.
        """
        assert expected_types, "Expected at least one type to expect"
        assert isinstance(operator_or_function, Operator) or operator is not None, (
            "Expecting arguments from an function requires call operator"
        )
        self.raise_for_enough_arguments(
            operator_or_function,
            inside_function,
            len(expected_types),
            operator=operator,
        )
        for expected_type in expected_types[::-1]:
            argument_type = self.pop_type_from_stack()
            if expected_type not in (T.ANY, argument_type):
                if isinstance(operator_or_function, Operator):
                    raise TypecheckInvalidOperatorArgumentTypeError(
                        operator=operator_or_function,
                        actual_type=argument_type,
                        expected_type=expected_type,
                    )
                raise TypecheckInvalidFunctionArgumentTypeError(
                    function=operator_or_function,
                    actual_type=argument_type,
                    expected_type=expected_type,
                )



"""General consts and types for registers and architecture (including FFI/ABI/IPC)."""

from typing import Literal

from gofra.parser.intrinsics import Intrinsic

CODEGEN_ENTRY_POINT_SYMBOL = "_start"
CODEGEN_GOFRA_CONTEXT_LABEL = ".L_%s_%s"

type CODEGEN_GOFRA_ON_STACK_OPERATIONS = Literal[
    "+",
    "-",
    "*",
    "//",
    "%",
    ">",
    ">=",
    "<",
    "<=",
    "==",
    "!=",
    "++",
    "--",
]

CODEGEN_INTRINSIC_TO_ASSEMBLY_OPS: dict[
    Intrinsic,
    CODEGEN_GOFRA_ON_STACK_OPERATIONS,
] = {
    Intrinsic.PLUS: "+",
    Intrinsic.MINUS: "-",
    Intrinsic.MULTIPLY: "*",
    Intrinsic.DIVIDE: "//",
    Intrinsic.MODULUS: "%",
    Intrinsic.INCREMENT: "++",
    Intrinsic.DECREMENT: "--",
    Intrinsic.NOT_EQUAL: "!=",
    Intrinsic.GREATER_EQUAL_THAN: ">=",
    Intrinsic.LESS_EQUAL_THAN: "<=",
    Intrinsic.LESS_THAN: "<",
    Intrinsic.GREATER_THAN: ">",
    Intrinsic.EQUAL: "==",
}



from collections.abc import Sequence

from gofra.exceptions import GofraError
from gofra.parser.functions.function import Function
from gofra.parser.operators import Operator

from .types import GofraType


class TypecheckInvalidOperatorArgumentTypeError(GofraError):
    def __init__(
        self,
        *args: object,
        expected_type: GofraType,
        actual_type: GofraType,
        operator: Operator,
    ) -> None:
        super().__init__(*args)
        self.expected_type = expected_type
        self.actual_type = actual_type
        self.operator = operator

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Expected {self.expected_type.name} but got {self.actual_type.name}
 for '{self.operator.token.text}' at {self.operator.token.location}

Did you miss the types?"""


class TypecheckInvalidFunctionArgumentTypeError(GofraError):
    def __init__(
        self,
        *args: object,
        expected_type: GofraType,
        actual_type: GofraType,
        function: Function,
    ) -> None:
        super().__init__(*args)
        self.expected_type = expected_type
        self.actual_type = actual_type
        self.function = function

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Expected {self.expected_type.name} but got {self.actual_type.name}
 for function '{self.function.name}' at {self.function.location}

Did you miss the types?"""


class TypecheckInvalidBinaryMathArithmeticsError(GofraError):
    def __init__(
        self,
        *args: object,
        actual_lhs_type: GofraType,
        actual_rhs_type: GofraType,
        operator: Operator,
    ) -> None:
        super().__init__(*args)
        self.actual_lhs_type = actual_lhs_type
        self.actual_rhs_type = actual_rhs_type
        self.operator = operator

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Binary math operator '{self.operator.token.text}' at {self.operator.token.location} expected both {GofraType.INTEGER.name} operands, 
but got {self.actual_lhs_type.name} on the left and {self.actual_rhs_type.name} on the right.

Expected contract: [{GofraType.INTEGER.name}, {GofraType.INTEGER.name}]
Actual contract: [{self.actual_lhs_type.name}, {self.actual_rhs_type.name}]

Pointer arithmetics disallowed within binary math operators!
Please use desired pointer-arithmetics safe operators!

Did you miss the types?"""


class TypecheckInvalidPointerArithmeticsError(GofraError):
    def __init__(
        self,
        *args: object,
        actual_lhs_type: GofraType,
        actual_rhs_type: GofraType,
        operator: Operator,
    ) -> None:
        super().__init__(*args)
        self.actual_lhs_type = actual_lhs_type
        self.actual_rhs_type = actual_rhs_type
        self.operator = operator

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Invalid pointer arithmetics for operator '{self.operator.token.text}' at {self.operator.token.location}

Expected contract: [{GofraType.POINTER.name}, {GofraType.INTEGER.name}]
Actual contract: [{self.actual_lhs_type.name}, {self.actual_rhs_type.name}]

Did you miss the types?"""


class TypecheckFunctionTypeContractOutViolatedError(GofraError):
    def __init__(
        self,
        *args: object,
        function: Function,
        type_stack: list[GofraType],
    ) -> None:
        super().__init__(*args)
        self.function = function
        self.type_stack = type_stack

    def __repr__(self) -> str:
        return f"""Type safety check error!

Function `{self.function.name}` at {self.function.location} has type contract out {self.function.type_contract_out}

But actual stack at the end is: {self.type_stack}"""


class TypecheckNotEnoughOperatorArgumentsError(GofraError):
    def __init__(
        self,
        *args: object,
        types_on_stack: Sequence[GofraType],
        required_args: int,
        operator: Operator,
    ) -> None:
        super().__init__(*args)
        self.types_on_stack = types_on_stack
        self.required_args = required_args
        self.operator = operator

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Expected {self.required_args} arguments on stack but got {len(self.types_on_stack)}
 for '{self.operator.token.text}' at {self.operator.token.location}

Did you miss some arguments?"""


class TypecheckNotEnoughFunctionArgumentsError(GofraError):
    def __init__(
        self,
        *args: object,
        types_on_stack: Sequence[GofraType],
        function: Function,
        callee_function: Function,
        called_from_operator: Operator,
    ) -> None:
        super().__init__(*args)
        self.types_on_stack = types_on_stack
        self.function = function
        self.callee_function = callee_function
        self.called_from_operator = called_from_operator

    def __repr__(self) -> str:
        return f"""Type safety check failed!

Expected {len(self.function.type_contract_in)} arguments on stack but got {len(self.types_on_stack)}
For function '{self.function.name}' defined at {self.function.location}
Called from function `{self.callee_function.name}` with `{self.called_from_operator.token.text}` at `{self.called_from_operator.token.location}`

Function type contract in: {self.function.type_contract_in}
But types on stack is: {self.types_on_stack}

Did you miss some arguments?"""


class TypecheckBlockStackMismatchError(GofraError):
    def __init__(
        self,
        *args: object,
        operator_begin: Operator,
        operator_end: Operator,
        stack_before_block: Sequence[GofraType],
        stack_after_block: Sequence[GofraType],
    ) -> None:
        super().__init__(*args)
        self.operator_begin = operator_begin
        self.operator_end = operator_end
        self.stack_before_block = stack_before_block
        self.stack_after_block = stack_after_block

    def __repr__(self) -> str:
        return f"""Stack mismatch!

Expected that `{self.operator_begin.token.text}` block at {self.operator_begin.token.location} will not modify stack state!
(Block ends with `{self.operator_end.token.text}` at {self.operator_end.token.location})

Before block stack types was: {self.stack_before_block}
After block stack types become: {self.stack_after_block}

Blocks should not modify stack state.
Please ensure that stacks are same!"""



"""Intermidiate representation for CLI.

Allows to view IR from CLI.
"""

from gofra.consts import GOFRA_ENTRY_POINT
from gofra.context import ProgramContext
from gofra.parser.functions.function import Function
from gofra.parser.intrinsics import Intrinsic
from gofra.parser.operators import Operator, OperatorType


def emit_ir_into_stdout(context: ProgramContext) -> None:
    """Display IR via stdout."""
    functions = {**context.functions, GOFRA_ENTRY_POINT: context.entry_point}
    for function in functions.values():
        emit_ir_function_signature(function, context.entry_point)
        context_block_shift = 0
        for operator in function.source:
            if operator.type in (OperatorType.DO, OperatorType.END):
                context_block_shift -= 1
            emit_ir_operator(operator, context_block_shift=context_block_shift)
            if operator.type in (OperatorType.DO, OperatorType.IF, OperatorType.WHILE):
                context_block_shift += 1


def emit_ir_operator(operator: Operator, context_block_shift: int) -> None:  # noqa: PLR0911
    shift = " " * (context_block_shift + 3)
    assert (  # noqa: PT018
        not operator.has_optimizations
        and not operator.infer_type_after_optimization
        and not operator.syscall_optimization_injected_args
        and not operator.syscall_optimization_omit_result
    ), "Optimizations are not implemented in IR representation"
    match operator.type:
        case OperatorType.PUSH_INTEGER:
            return print(f"{shift}PUSH {operator.operand}")
        case OperatorType.INTRINSIC:
            assert isinstance(operator.operand, Intrinsic)
            return print(f"{shift}{operator.operand.name}")
        case OperatorType.PUSH_MEMORY_POINTER:
            return print(f"{shift}PUSH_MEM '{operator.operand}'")
        case OperatorType.WHILE | OperatorType.IF:
            return print(f"{shift}{operator.type.name}" + "{")
        case OperatorType.DO:
            return print(f"{shift}" + "}" + f"{operator.type.name}" + "{")
        case OperatorType.END:
            return print(f"{shift}" + "}")
        case OperatorType.FUNCTION_CALL:
            return print(f"{shift}{operator.operand}()")
        case _:
            return print(f"{shift}{operator.type.name}<{operator.operand}>")


def emit_ir_function_signature(function: Function, entry_point: Function) -> None:
    if function.is_externally_defined:
        print(f"[external function symbol '{function.name}'", end=" ")
        print(f"({function.type_contract_in} -> {function.type_contract_out})")
        return
    if function == entry_point:
        print(f"[entry point symbol '{function.name}']")
        return
    print(f"[function symbol '{function.name}'", end=" ")
    print(f"({function.type_contract_in} -> {function.type_contract_out})", end=" ")
    print(f"(global={function.is_global_linker_symbol})]")



from abc import abstractmethod


class GofraError(Exception):
    """Parent for all Gofra errors (exceptions)."""

    @abstractmethod
    def __repr__(self) -> str:
        return f"Some internal error occurred ({super().__repr__()}), that is currently not documented"



"""Preprocessor which resolves includes, CTE/macros and other from given lexer.

Simply, wraps an lexer into another `lexer` and preprocess on the fly.
"""

from .preprocessor import preprocess_file

__all__ = ("preprocess_file",)



"""Parser for function definitions in Gofra language.

Functions is an block like:
{inline|extern...} func function_name[signature_types,...] return_type

For example:
```
func my_func[] void ... end
inline func do_something[int, int] int ... end
extern func puts[ptr] int
```

Extern functions cannot have a body so they do not have `end` block (assuming that - does not have any body block).
"""

from gofra.lexer import Token
from gofra.lexer.keywords import Keyword
from gofra.lexer.tokens import TokenType
from gofra.parser._context import ParserContext
from gofra.typecheck.types import WORD_TO_GOFRA_TYPE, GofraType

from .exceptions import (
    ParserExpectedFunctionAfterFunctionModifiersError,
    ParserExpectedFunctionKeywordError,
    ParserExpectedFunctionReturnTypeError,
    ParserFunctionInvalidTypeError,
    ParserFunctionIsBothInlineAndExternalError,
    ParserFunctionModifierReappliedError,
    ParserFunctionNoNameError,
)


def consume_function_definition(
    context: ParserContext,
    token: Token,
) -> tuple[Token, str, list[GofraType], list[GofraType], bool, bool, bool]:
    token, (modifier_is_inline, modifier_is_extern, modifier_is_global) = (
        consume_function_modifiers(
            context,
            token,
        )
    )
    function_name, type_contract_in, type_contract_out = consume_function_signature(
        context,
        token,
    )

    return (
        token,
        function_name,
        type_contract_in,
        type_contract_out,
        modifier_is_inline,
        modifier_is_extern,
        modifier_is_global,
    )


def consume_function_modifiers(
    context: ParserContext,
    token: Token,
) -> tuple[Token, tuple[bool, bool, bool]]:
    """Consume parser context assuming given token is last popped, and it is a function modifier (or base function).

    Accepts `inline`, `extern`, `function` keywords as tokens.
    Returns the last token (function name) and a tuple of flags (is_inline, is_extern).
    """
    # Function modifier parsing must be started from modifier or start
    assert token.type == TokenType.KEYWORD
    assert token.value in (
        Keyword.INLINE,
        Keyword.EXTERN,
        Keyword.FUNCTION,
        Keyword.GLOBAL,
    )

    mark_is_extern = False
    mark_is_inline = False
    mark_is_global = False

    next_token = token
    while next_token:
        if next_token.type != TokenType.KEYWORD:
            raise ParserExpectedFunctionKeywordError(token=next_token)

        match next_token.value:
            case Keyword.INLINE:
                if mark_is_inline:
                    raise ParserFunctionModifierReappliedError(
                        modifier_token=next_token,
                    )
                mark_is_inline = True
            case Keyword.EXTERN:
                if mark_is_extern:
                    raise ParserFunctionModifierReappliedError(
                        modifier_token=next_token,
                    )
                mark_is_extern = True
            case Keyword.FUNCTION:
                break
            case Keyword.GLOBAL:
                if mark_is_global:
                    raise ParserFunctionModifierReappliedError(
                        modifier_token=next_token,
                    )
                mark_is_global = True
            case _:
                raise ParserExpectedFunctionKeywordError(token=next_token)

        if mark_is_extern and mark_is_inline:
            raise ParserFunctionIsBothInlineAndExternalError(
                modifier_token=next_token,
            )
        next_token = next(context.tokenizer, None)

    if not next_token:
        raise NotImplementedError
    if next_token.type != TokenType.KEYWORD or next_token.value != Keyword.FUNCTION:
        raise ParserExpectedFunctionAfterFunctionModifiersError(modifier_token=token)

    return next_token, (mark_is_inline, mark_is_extern, mark_is_global)


def consume_function_signature(
    context: ParserContext,
    token: Token,
) -> tuple[str, list[GofraType], list[GofraType]]:
    """Consume parser context into function signature assuming given token is `function` keyword.

    Returns function name and signature types (`in` and `out).
    """

    sig_token = next(context.tokenizer, None)
    if not sig_token:
        raise ParserExpectedFunctionReturnTypeError(
            definition_token=token,
        )
    if sig_token.type != TokenType.WORD:
        raise ValueError

    type_contract_out = _parse_function_type_contract(
        token=sig_token,
        contract=(
            sig_token.text
            if sig_token.text.startswith("[") and sig_token.text.endswith("]")
            else f"[{sig_token.text}]"
        ),
    )

    signature_token = next(context.tokenizer, None)
    if not signature_token:
        raise ParserFunctionNoNameError(token=token)

    signature = signature_token.text

    # Assume for the first time that function has no input contract
    function_name = signature
    type_contract_in = []

    if "[" in signature and "]" in signature:
        function_name = function_name.split("[")[0].strip()
        contract = signature.split("[")[1].strip()[:-1]

        type_contract_in = _parse_function_type_contract(
            token=signature_token,
            contract=f"[{contract}]",
        )

    return function_name, type_contract_in, type_contract_out


def _parse_function_type_contract(token: Token, contract: str) -> list[GofraType]:
    """Parse function type contract from string.

    Expected contract to be like: `[ptr,int,int]`
    """
    assert contract.startswith("[")
    assert contract.endswith("]")

    raw_contract_types = [t.strip() for t in contract[1:-1].split(",") if t != ""]

    for raw_contract_type in raw_contract_types:
        if raw_contract_type not in WORD_TO_GOFRA_TYPE:
            raise ParserFunctionInvalidTypeError(
                type_token=token,
                requested_type=raw_contract_type,
            )

    return [
        WORD_TO_GOFRA_TYPE[raw_contract_type]
        for raw_contract_type in raw_contract_types
        if raw_contract_type != ""
        and WORD_TO_GOFRA_TYPE[raw_contract_type] != GofraType.VOID
    ]


"""
def _consume_function_definition_name_with_types(
    context: ParserContext,
    token: Token,
) -> tuple[GofraType, str, list[GofraType]]:


    if function_name in context.macros:
        macro_reference = context.macros[function_name]
        raise ParserExternRedefinesMacroError(
            redefine_extern_function_name_token=function_sig_def_token,
            original_macro_name=macro_reference.name,
            original_macro_location=macro_reference.location,
        )

    if function_name in (WORD_TO_INTRINSIC.keys() | WORD_TO_KEYWORD.keys()):
        raise ParserExternRedefinesLanguageDefinitionError(
            extern_token=function_sig_def_token,
            extern_function_name=function_name,
        )

    return return_type, function_name, signature_types
"""



from pathlib import Path

from gofra.lexer.tokens import Token
from gofra.preprocessor.exceptions import PreprocessorError


class PreprocessorIncludeFileNotFoundError(PreprocessorError):
    def __init__(self, *args: object, include_token: Token, include_path: Path) -> None:
        super().__init__(*args)
        self.include_token = include_token
        self.include_path = include_path

    def __repr__(self) -> str:
        return f"""Unable to include file '{self.include_path}' requested at {self.include_token.location}

File does not exists!

Please check that this file exists, or try updating include directory paths.
Requested path directly resolves to '{self.include_path.resolve()}'
(Does not includes import directory traverses)

Did you supplied wrong name/path?"""


class PreprocessorIncludeNoPathError(PreprocessorError):
    def __init__(self, *args: object, include_token: Token) -> None:
        super().__init__(*args)
        self.include_token = include_token

    def __repr__(self) -> str:
        return f"""'include' has no path at {self.include_token.location}!

Expected there will be include path after 'include' as string.

Did you forgot to add path?"""


class PreprocessorIncludeNonStringNameError(PreprocessorError):
    def __init__(self, *args: object, include_path_token: Token) -> None:
        super().__init__(*args)
        self.include_path_token = include_path_token

    def __repr__(self) -> str:
        return f"""Invalid include path type at {self.include_path_token.location}!

Expected include path as string with quotes.

Did you forgot to add quotes?"""


class PreprocessorIncludeCurrentFileError(PreprocessorError):
    def __init__(self, *args: object, include_path_token: Token) -> None:
        super().__init__(*args)
        self.include_path_token = include_path_token

    def __repr__(self) -> str:
        return f"""Tried to include current file within include at {self.include_path_token.location}!
Including self is prohibited and will lead to no actions, so please remove self-import

Did you mistyped import path?"""



from collections.abc import Generator
from contextlib import contextmanager

from gofra.exceptions import GofraError

from .output import cli_message

DEBUG_UNWRAP_INTERNAL_ERRORS = True


@contextmanager
def cli_gofra_error_handler() -> Generator[None]:
    """Wrap function to properly emit Gofra internal errors."""
    try:
        yield
    except GofraError as ge:
        if DEBUG_UNWRAP_INTERNAL_ERRORS:
            cli_message("ERROR", repr(ge))
        else:
            raise



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from gofra.parser._context import ParserContext
    from gofra.parser.functions import Function


@dataclass(frozen=False)
class ProgramContext:
    """Context for program which only required for internal usages.

    Acquired from parser context and mutable within next stages.
    """

    functions: MutableMapping[str, Function]
    memories: MutableMapping[str, int]
    entry_point: Function

    @staticmethod
    def from_parser_context(
        parser_context: ParserContext,
        entry_point: Function,
    ) -> ProgramContext:
        return ProgramContext(
            functions=parser_context.functions,
            memories=parser_context.memories,
            entry_point=entry_point,
        )



"""Consts and types related to AMD64 registers and architecture (including FFI/ABI/IPC)."""

from __future__ import annotations

from typing import Literal

####
# Bare AMD64 related
####

# Registers specification for AMD64
# Skips some of registers due to currently being unused
type AMD64_GP_REGISTERS = Literal[
    "rax",
    "rbx",
    "rdi",
    "rsi",
    "rcx",
    "rdx",
    "r10",
    "r8",
    "r9",
]


####
# Linux related
####

# Epilogue
AMD64_LINUX_SYSCALL_NUMBER_REGISTER: AMD64_GP_REGISTERS = "rax"
AMD64_LINUX_SYSCALL_ARGUMENTS_REGISTERS: tuple[AMD64_GP_REGISTERS, ...] = (
    "rdi",
    "rsi",
    "rdx",
    "r10",
    "r8",
    "r9",
)
AMD64_LINUX_ABI_RETVAL_REGISTER: AMD64_GP_REGISTERS = "rax"
AMD64_LINUX_EPILOGUE_EXIT_CODE = 0
AMD64_LINUX_EPILOGUE_EXIT_SYSCALL_NUMBER = 60
AMD64_LINUX_ABI_ARGUMENTS_REGISTERS: tuple[AMD64_GP_REGISTERS, ...] = (
    "rdi",
    "rsi",
    "rdx",
    "rcx",
    "r8",
    "r9",
)



from __future__ import annotations

from difflib import get_close_matches
from typing import TYPE_CHECKING, assert_never

from gofra.lexer import (
    Keyword,
    Token,
    TokenType,
)
from gofra.lexer.keywords import KEYWORD_TO_NAME, WORD_TO_KEYWORD
from gofra.parser.functions.parser import consume_function_definition
from gofra.parser.validator import validate_and_pop_entry_point

from ._context import ParserContext
from .exceptions import (
    ParserDirtyNonPreprocessedTokenError,
    ParserEmptyIfBodyError,
    ParserEndAfterWhileError,
    ParserEndWithoutContextError,
    ParserExhaustiveContextStackError,
    ParserNoWhileBeforeDoError,
    ParserNoWhileConditionOperatorsError,
    ParserUnfinishedIfBlockError,
    ParserUnfinishedWhileDoBlockError,
    ParserUnknownWordError,
)
from .intrinsics import WORD_TO_INTRINSIC
from .operators import OperatorType

if TYPE_CHECKING:
    from collections.abc import Generator

    from gofra.parser.functions import Function


def parse_file(tokenizer: Generator[Token]) -> tuple[ParserContext, Function]:
    """Load file for parsing into operators."""
    context = _parse_from_context_into_operators(
        context=ParserContext(
            is_top_level=True,
            tokenizer=tokenizer,
            functions={},
            memories={},
        ),
    )

    assert context.is_top_level
    assert not context.operators

    entry_point = validate_and_pop_entry_point(context)
    return context, entry_point


def _parse_from_context_into_operators(context: ParserContext) -> ParserContext:
    """Consumes token stream into language operators."""
    while token := next(context.tokenizer, None):
        _consume_token_for_parsing(
            token=token,
            context=context,
        )

    if context.context_stack:
        _, unclosed_operator = context.pop_context_stack()
        match unclosed_operator.type:
            case OperatorType.DO | OperatorType.WHILE:
                raise ParserUnfinishedWhileDoBlockError(token=unclosed_operator.token)
            case OperatorType.IF:
                raise ParserUnfinishedIfBlockError(if_token=unclosed_operator.token)
            case _:
                raise ParserExhaustiveContextStackError

    return context


def _consume_token_for_parsing(token: Token, context: ParserContext) -> None:
    match token.type:
        case TokenType.INTEGER | TokenType.CHARACTER:
            return _push_integer_operator(context, token)
        case TokenType.STRING:
            return _push_string_operator(context, token)
        case TokenType.WORD:
            if _try_unpack_macro_or_inline_function_from_token(context, token):
                return None

            if _try_unpack_memory_reference_from_token(context, token):
                return None

            if _try_push_intrinsic_operator(context, token):
                return None

            raise ParserUnknownWordError(
                word_token=token,
                macro_names=context.functions.keys(),
                best_match=_best_match_for_word(context, token.text),
            )
        case TokenType.KEYWORD:
            return _consume_keyword_token(context, token)


def _best_match_for_word(context: ParserContext, word: str) -> str | None:
    matches = get_close_matches(
        word,
        WORD_TO_INTRINSIC.keys() | context.functions.keys(),
    )
    return matches[0] if matches else None


def _consume_keyword_token(context: ParserContext, token: Token) -> None:
    assert isinstance(token.value, Keyword)
    TOP_LEVEL_KEYWORD = (  # noqa: N806
        Keyword.INLINE,
        Keyword.EXTERN,
        Keyword.FUNCTION,
        Keyword.GLOBAL,
        Keyword.MEMORY,
        # TODO(@kirillzhosul): Remove reference
        Keyword.PP_ENDIF,
        Keyword.PP_IFDEF,
        Keyword.PP_INCLUDE,
        Keyword.PP_MACRO,
    )
    if context.is_top_level and token.value not in (*TOP_LEVEL_KEYWORD, Keyword.END):
        msg = f"{token.value.name} expected to be not at top level! (temp-assert)"
        raise NotImplementedError(msg)
    if not context.is_top_level and token.value in TOP_LEVEL_KEYWORD:
        msg = f"{token.value.name} expected to be at top level! (temp-assert)"
        raise NotImplementedError(msg)
    match token.value:
        case Keyword.IF | Keyword.DO | Keyword.WHILE | Keyword.END:
            return _consume_conditional_keyword_from_token(context, token)
        case (
            Keyword.PP_INCLUDE | Keyword.PP_MACRO | Keyword.PP_IFDEF | Keyword.PP_ENDIF
        ):
            raise ParserDirtyNonPreprocessedTokenError(token=token)
        case Keyword.INLINE | Keyword.EXTERN | Keyword.FUNCTION | Keyword.GLOBAL:
            return _unpack_function_definition_from_token(context, token)
        case Keyword.FUNCTION_CALL:
            return _unpack_function_call_from_token(context, token)
        case Keyword.FUNCTION_RETURN:
            return context.push_new_operator(
                OperatorType.FUNCTION_RETURN,
                token,
                None,
                is_contextual=False,
            )
        case Keyword.MEMORY:
            return _unpack_memory_segment_from_token(context, token)
        case _:
            assert_never(token.value)


def _unpack_memory_segment_from_token(context: ParserContext, token: Token) -> None:
    memory_segment_name = next(context.tokenizer, None)
    if not memory_segment_name:
        raise NotImplementedError
    if memory_segment_name.type != TokenType.WORD:
        raise NotImplementedError
    assert isinstance(memory_segment_name.value, str)
    memory_segment_size = next(context.tokenizer, None)
    if not memory_segment_size:
        raise NotImplementedError
    if memory_segment_size.type != TokenType.INTEGER:
        raise NotImplementedError
    assert isinstance(memory_segment_size.value, int)

    # This is an definition only so we dont acquire reference/pointer
    context.memories[memory_segment_name.value] = memory_segment_size.value


def _unpack_function_call_from_token(context: ParserContext, token: Token) -> None:
    extern_call_name_token = next(context.tokenizer, None)
    if not extern_call_name_token:
        raise NotImplementedError
    extern_call_name = extern_call_name_token.text

    if extern_call_name_token.type != TokenType.WORD:
        raise NotImplementedError

    target_function = context.functions.get(extern_call_name)
    if not target_function:
        raise NotImplementedError(f"Unknown function {extern_call_name}")

    if target_function.emit_inline_body:
        _try_unpack_macro_or_inline_function_from_token(context, extern_call_name_token)
        return

    context.push_new_operator(
        OperatorType.FUNCTION_CALL,
        token,
        extern_call_name,
        is_contextual=False,
    )


def _unpack_function_definition_from_token(
    context: ParserContext,
    token: Token,
) -> None:
    definition = consume_function_definition(context, token)
    (
        token,
        function_name,
        type_contract_in,
        type_contract_out,
        modifier_is_inline,
        modifier_is_extern,
        modifier_is_global,
    ) = definition

    if modifier_is_extern:
        if len(type_contract_out) > 1:
            msg = "Extern functions cannot have stack type contract consider using C FFI ABI"
            raise NotImplementedError(msg)
        context.new_function(
            from_token=token,
            name=function_name,
            type_contract_in=type_contract_in,
            type_contract_out=type_contract_out,
            emit_inline_body=modifier_is_inline,
            is_externally_defined=modifier_is_extern,
            is_global_linker_symbol=modifier_is_global,
            source=[],
        )
        return

    opened_context_blocks = 0
    function_was_closed = False

    context_keywords = (Keyword.IF, Keyword.DO)
    end_keyword_text = KEYWORD_TO_NAME[Keyword.END]

    original_token = token
    function_body_tokens: list[Token] = []
    while func_token := next(context.tokenizer, None):
        if func_token.type != TokenType.KEYWORD:
            function_body_tokens.append(func_token)
            continue

        if func_token.text == end_keyword_text:
            if opened_context_blocks <= 0:
                function_was_closed = True
                break
            opened_context_blocks -= 1

        is_context_keyword = WORD_TO_KEYWORD[func_token.text] in context_keywords
        if is_context_keyword:
            opened_context_blocks += 1

        function_body_tokens.append(func_token)

    if not func_token:
        raise NotImplementedError
    if not function_was_closed:
        raise ValueError(original_token)

    new_context = ParserContext(
        is_top_level=False,
        tokenizer=(t for t in function_body_tokens),
        functions=context.functions,
        memories=context.memories,
    )
    context.new_function(
        from_token=func_token,
        name=function_name,
        type_contract_in=type_contract_in,
        type_contract_out=type_contract_out,
        emit_inline_body=modifier_is_inline,
        is_externally_defined=modifier_is_extern,
        is_global_linker_symbol=modifier_is_global,
        source=_parse_from_context_into_operators(context=new_context).operators,
    )


def _consume_conditional_keyword_from_token(
    context: ParserContext,
    token: Token,
) -> None:
    assert isinstance(token.value, Keyword)
    match token.value:
        case Keyword.IF:
            return context.push_new_operator(
                type=OperatorType.IF,
                token=token,
                operand=None,
                is_contextual=True,
            )
        case Keyword.DO:
            if not context.has_context_stack():
                raise ParserNoWhileBeforeDoError(do_token=token)

            operator_while_idx, context_while = context.pop_context_stack()
            if context_while.type != OperatorType.WHILE:
                raise ParserNoWhileBeforeDoError(do_token=token)

            while_condition_len = context.current_operator - operator_while_idx - 1
            if while_condition_len == 0:
                raise ParserNoWhileConditionOperatorsError(
                    while_token=context_while.token,
                )

            operator = context.push_new_operator(
                type=OperatorType.DO,
                token=token,
                operand=None,
                is_contextual=True,
            )
            context.operators[-1].jumps_to_operator_idx = operator_while_idx
            return operator
        case Keyword.WHILE:
            return context.push_new_operator(
                type=OperatorType.WHILE,
                token=token,
                operand=None,
                is_contextual=True,
            )
        case Keyword.END:
            if not context.has_context_stack():
                raise ParserEndWithoutContextError(end_token=token)

            context_operator_idx, context_operator = context.pop_context_stack()

            context.push_new_operator(
                type=OperatorType.END,
                token=token,
                operand=None,
                is_contextual=False,
            )
            prev_context_jumps_at = context_operator.jumps_to_operator_idx
            context_operator.jumps_to_operator_idx = context.current_operator - 1

            match context_operator.type:
                case OperatorType.DO:
                    context.operators[-1].jumps_to_operator_idx = prev_context_jumps_at
                case OperatorType.IF:
                    if_body_size = context.current_operator - context_operator_idx - 2
                    if if_body_size == 0:
                        raise ParserEmptyIfBodyError(if_token=context_operator.token)
                case OperatorType.WHILE:
                    raise ParserEndAfterWhileError(end_token=token)
                case _:
                    raise AssertionError

            return None
        case _:
            raise AssertionError


def _try_unpack_memory_reference_from_token(
    context: ParserContext,
    token: Token,
) -> bool:
    assert token.type == TokenType.WORD

    memory_name = token.text
    if memory_name not in context.memories:
        return False
    context.push_new_operator(
        type=OperatorType.PUSH_MEMORY_POINTER,
        token=token,
        operand=memory_name,
        is_contextual=False,
    )
    return True


def _try_unpack_macro_or_inline_function_from_token(
    context: ParserContext,
    token: Token,
) -> bool:
    assert token.type == TokenType.WORD

    inline_block = context.functions.get(token.text, None)

    if inline_block:
        if not inline_block.emit_inline_body or inline_block.is_externally_defined:
            raise NotImplementedError(
                f"use `call` to call an function, obtaining an function is not implemented yet {token.location}"
            )

        context.expand_from_inline_block(inline_block)

    return bool(inline_block)


def _push_string_operator(context: ParserContext, token: Token) -> None:
    assert isinstance(token.value, str)
    context.push_new_operator(
        type=OperatorType.PUSH_STRING,
        token=token,
        operand=token.value,
        is_contextual=False,
    )


def _push_integer_operator(context: ParserContext, token: Token) -> None:
    assert isinstance(token.value, int)
    context.push_new_operator(
        type=OperatorType.PUSH_INTEGER,
        token=token,
        operand=token.value,
        is_contextual=False,
    )


def _try_push_intrinsic_operator(context: ParserContext, token: Token) -> bool:
    assert isinstance(token.value, str)
    intrinsic = WORD_TO_INTRINSIC.get(token.value)

    if intrinsic is None:
        return False

    context.push_new_operator(
        type=OperatorType.INTRINSIC,
        token=token,
        operand=intrinsic,
        is_contextual=False,
    )
    return True



from __future__ import annotations

from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class CLIArguments:
    """Arguments from argument parser provided for whole Gofra testkit process."""

    directory: Path
    build_cache_dir: Path


def parse_cli_arguments() -> CLIArguments:
    """Parse CLI arguments from argparse into custom DTO."""
    args = _construct_argument_parser().parse_args()

    return CLIArguments(
        directory=Path(args.directory),
        build_cache_dir=Path(args.cache_dir),
    )


def _construct_argument_parser() -> ArgumentParser:
    """Get argument parser instance to parse incoming arguments."""
    parser = ArgumentParser(
        description="Gofra Testkit - CLI for testing internals of Gofra programming language",
        add_help=True,
    )

    parser.add_argument(
        "--directory",
        "-d",
        default="./",
        required=False,
        help="Directory from which to search test files for runner. Defaults to `./`",
    )
    parser.add_argument(
        "--cache-dir",
        "-cd",
        type=str,
        default="./.build",
        required=False,
        help="Path to directory where to store cache (defaults `./.build`)",
    )
    return parser



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.lexer.keywords import Keyword
from gofra.lexer.tokens import Token, TokenType

from .exceptions import (
    PreprocessorConditionalConsumeUntilEndifContextSwitchError,
    PreprocessorConditionalNoMacroNameError,
)

if TYPE_CHECKING:
    from gofra.preprocessor._state import PreprocessorState
    from gofra.preprocessor.macros.container import Macro


def resolve_conditional_block_from_token(
    token: Token,
    state: PreprocessorState,
) -> None:
    macro = _consume_macro_from_token(token, state)
    match token:
        case Token(type=TokenType.KEYWORD, value=Keyword.PP_IFDEF):
            if not macro:
                _consume_until_endif(token, state)
        case _:
            msg = "Expected resolve conditional block to receive an known preprocessor keyword token"
            raise AssertionError(msg)


def _consume_until_endif(from_token: Token, state: PreprocessorState) -> None:
    while token := next(state.tokenizer, None):
        if token.type != TokenType.KEYWORD:
            continue
        if token.value == Keyword.PP_ENDIF:
            return
        if token.location.filepath != from_token.location.filepath:
            raise PreprocessorConditionalConsumeUntilEndifContextSwitchError(
                conditional_token=from_token,
            )


def _consume_macro_from_token(token: Token, state: PreprocessorState) -> Macro | None:
    macro_name = next(state.tokenizer, None)
    if not macro_name:
        raise PreprocessorConditionalNoMacroNameError(conditional_token=token)
    if macro_name.type != TokenType.WORD:
        raise ValueError
    assert isinstance(macro_name.value, str)
    return state.macros.get(macro_name.value)



"""Function object as DTO and operations between stages."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from gofra.lexer.tokens import TokenLocation
    from gofra.parser.operators import Operator
    from gofra.typecheck.types import GofraType


# Both input and output type contracts are the same
type FunctionTypeContract = Sequence[GofraType]


class Function:
    """Language level function definition.

    Function is a sequence of operators (source body) that can be called by name

    For base functions it will define an executable region of code (real function) at code generation step
    `Extern` functions has no body and just an link for an external function
    `Inline` functions will be expanded within call (macros expansion) and will not be called like normal function

    Functions in Gofra are not first class functions
    They are not passed as arguments and are not returned from other functions
    They are just a sequence of operators (executable region of source body)
    """

    # Location of the function definition
    # Should not be used in any way for code generation or type checking
    # This is only for debugging purposes
    location: TokenLocation

    # Function will be called by that name
    # For extern functions, on different backends conventions is quite different
    # So on MacOS it will be prefixed with `_` and on Linux it will not
    name: str

    # Actual executable region that this function contains
    # If this is extern function will always be empty
    # If this is inline function will expand like there were no function
    source: Sequence[Operator]

    # Type contract `in -> out`
    # Specifies which types this function accepts and which - returns
    # Code generator will generate desired wrapper around that contract system
    # Type checker also validate usage of function so it does not be used with wrong contract
    type_contract_in: FunctionTypeContract
    type_contract_out: FunctionTypeContract

    # If true `function call` rather than proceeding into `jumping` (`calling`) into that function
    # just inject (emit) body of the function inside call target location (expand body of the function from call)
    # affects code generator so it wont generate native function block and will not use jumps
    # TODO(@kirillzhosul): Current implementation pollutes original definition source (expand whole operator/token sequence)
    emit_inline_body: bool

    # If true, function must have empty body as it is located somewhere else and only available after assembler step
    # Extern functions mostly are `C` functions (system/user defined) and call to them does jumps inside linked source binaries (dynamic libraries)
    # Code generator takes care of that calls and specifies extern function requests if needed (aggregates them for that)
    is_externally_defined: bool

    is_global_linker_symbol: bool = False

    def __init__(  # noqa: PLR0913
        self,
        *,
        location: TokenLocation,
        name: str,
        source: Sequence[Operator],
        type_contract_in: FunctionTypeContract,
        type_contract_out: FunctionTypeContract,
        emit_inline_body: bool,
        is_externally_defined: bool,
        is_global_linker_symbol: bool,
    ) -> None:
        self.location = location
        self.name = name
        self.source = source
        self.type_contract_in = type_contract_in
        self.type_contract_out = type_contract_out
        self.emit_inline_body = emit_inline_body
        self.is_externally_defined = is_externally_defined
        self.is_global_linker_symbol = is_global_linker_symbol
        self._validate()

    def _validate(self) -> None:
        """Validate internally that function is correct.

        Actual error handling should be before that and function object should not be created.
        """
        if self.is_externally_defined:
            if self.emit_inline_body:
                msg = "Functions cannot be both marked as `inline` and `external`"
                raise ValueError(msg)
            if self.source:
                msg = "Functions that marked as `external` cannot have an body!"
                raise ValueError(msg)
            if self.is_global_linker_symbol:
                msg = "Functions that marked as `global` cannot have be external!"
                raise ValueError(msg)

        if not self.is_externally_defined and not self.source:
            msg = "Functions that not marked as `external` must have an body!"
            raise ValueError(msg)

    def has_executable_body(self) -> bool:
        return not self.emit_inline_body and not self.is_externally_defined

    def abi_ffi_push_retval_onto_stack(self) -> bool:
        return self.is_externally_defined and bool(self.type_contract_out)



"""Codegen package responsive for generation of code for assembling.

Provides different backends for OSxARCH pair
"""

from .generator import generate_code_for_assembler

__all__ = ["generate_code_for_assembler"]



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.lexer._state import LexerState
from gofra.lexer.exceptions import (
    LexerAmbigiousHexadecimalAlphabetError,
    LexerEmptyCharacterError,
    LexerExcessiveCharacterLengthError,
    LexerUnclosedCharacterQuoteError,
    LexerUnclosedStringQuoteError,
)
from gofra.lexer.helpers import (
    CHARACTER_QUOTE,
    HEXADECIMAL_MARK,
    SINGLE_LINE_COMMENT,
    STRING_QUOTE,
    find_string_end,
    find_word_end,
    find_word_start,
    is_valid_hexadecimal,
    is_valid_integer,
    unescape_string,
)
from gofra.lexer.io import open_source_file_line_stream
from gofra.lexer.keywords import WORD_TO_KEYWORD
from gofra.lexer.tokens import Token, TokenLocation, TokenType

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path


def tokenize_file(path: Path) -> Generator[Token]:
    """Open given file and stream lexical tokens via generator (perform lexical analysis).

    :returns tokenizer: Generator of tokens, in order from top to bottom of an file (default order)
    """
    state = LexerState(path=path)

    for row, line in enumerate(open_source_file_line_stream(path), start=0):
        state.line = line

        state.row = row
        state.col = find_word_start(line, 0)

        col_ends_at = len(state.line)
        while state.col < col_ends_at and (token := _tokenize_line_next_token(state)):
            yield token


def _tokenize_line_next_token(state: LexerState) -> Token | None:
    """Acquire token from current state and modify it to apply next tokenize, or None if reached end (EOL/EOF).

    :returns token: Token from state or None if state must go to the next line (no tokens).
    """
    # First symbol of an token, as we exclussivly tokenize character/string tokens due to their fixed/variadic length.
    symbol = state.line[state.col]

    if symbol == CHARACTER_QUOTE:
        state.col += len(CHARACTER_QUOTE)
        return _tokenize_character_symbol(state)

    if symbol == STRING_QUOTE:
        state.col += len(STRING_QUOTE)
        return _tokenize_string_symbol(state)

    return _tokenize_word_into_token(state)


def _tokenize_string_symbol(state: LexerState) -> Token:
    """Tokenize string symbol into an string."""
    location = state.current_location()

    if (ends_at := find_string_end(state.line, state.col)) is None:
        raise LexerUnclosedStringQuoteError(open_quote_location=location)

    string_unescaped_quoted = state.line[state.col - 1 : ends_at]
    # Unescaped version at current moment is being unused, and probably may be removeD
    string_text = unescape_string(string_unescaped_quoted.strip(STRING_QUOTE))

    state.col = find_word_start(state.line, ends_at)
    return Token(
        type=TokenType.STRING,
        text=string_unescaped_quoted,
        value=string_text,
        location=location,
    )


def _tokenize_character_symbol(state: LexerState) -> Token:
    """Tokenize character symbol into an character."""
    location = state.current_location()

    character_ends_at = state.line.find(CHARACTER_QUOTE, state.col)
    if character_ends_at == -1:
        raise LexerUnclosedCharacterQuoteError(open_quote_location=location)

    character_unescaped_quoted = state.line[state.col - 1 : character_ends_at + 1]
    character = unescape_string(character_unescaped_quoted[1:-1])

    character_len = len(character)
    if character_len == 0:
        raise LexerEmptyCharacterError(open_quote_location=location)
    if character_len != 1:
        raise LexerExcessiveCharacterLengthError(
            excess_begins_at=location,
            excess_by_count=character_len,
        )

    state.col = find_word_start(state.line, character_ends_at + 1)
    return Token(
        type=TokenType.CHARACTER,
        text=character_unescaped_quoted,
        value=ord(character),
        location=location,
    )


def _try_tokenize_numerable_into_token(
    word: str,
    location: TokenLocation,
) -> Token | None:
    """Try to consume given word into number token (integer / float) or return nothing.

    TODO(@kirillzhosul): Review introduction of an unary negation operation (includes negation of non numeric values)
    TODO(@kirillzhosul): Introduce float/decimal numbers
    TODO(@kirillzhosul): Introduce binary base for numeric values
    """
    base = 0
    sign = (1, -1)[word.startswith("-")]
    word = word.removeprefix("-")

    if is_valid_integer(word):
        base = 10
    elif word.startswith(HEXADECIMAL_MARK):
        base = 16
        if not is_valid_hexadecimal(word):
            raise LexerAmbigiousHexadecimalAlphabetError(
                hexadecimal_raw=word,
                number_location=location,
            )

    if not base:
        return None

    return Token(
        type=TokenType.INTEGER,
        text=word,
        value=int(word, base) * sign,
        location=location,
    )


def _tokenize_word_or_keyword_into_token(word: str, location: TokenLocation) -> Token:
    """Tokenize base word symbol into an word or keyword."""
    if keyword := WORD_TO_KEYWORD.get(word):
        return Token(
            type=TokenType.KEYWORD,
            text=word,
            location=location,
            value=keyword,
        )

    return Token(
        type=TokenType.WORD,
        text=word,
        location=location,
        value=word,
    )


def _tokenize_word_into_token(state: LexerState) -> Token | None:
    """Consumes `word` (e.g can be also an number) into token (non-string and non-character as must be handled above).

    :returns token: Token or None if should skip this line due to EOL as comment mark found.
    """
    word_ends_at = find_word_end(state.line, state.col)
    word = state.line[state.col : word_ends_at]

    if word.startswith(SINGLE_LINE_COMMENT):
        # Next words were prefixed with comment mark - skip whole line.
        return None

    # Remember current location and jump to next word for next() call.
    location = state.current_location()
    state.col = find_word_start(state.line, word_ends_at)

    if token := _try_tokenize_numerable_into_token(word, location):
        return token

    return _tokenize_word_or_keyword_into_token(word, location)



"""Lexer package that used to lex source text into tokens (e.g tokenization).

Lexer takes an input file and splits it content into tokens.

Tokens and lexer implemented a bit odd (https://en.wikipedia.org/wiki/Lexical_analysis)
Some token types are misleading, this code does not always imply conventions.
"""

from .exceptions import LexerError
from .keywords import Keyword
from .lexer import tokenize_file
from .tokens import Token, TokenType

__all__ = [
    "Keyword",
    "LexerError",
    "Token",
    "TokenType",
    "tokenize_file",
]



from sys import stderr, stdout
from typing import Literal

type MessageLevel = Literal["INFO", "ERROR", "WARNING", "SUCCESS"]


class CLIColor:
    RED = "\033[91m"
    GREEN = "\033[92m"
    RESET = "\033[0m"
    YELLOW = "\033[93m"


def cli_message(level: MessageLevel, text: str, *, verbose: bool = True) -> None:
    """Emit an message to CLI user with given level, applying FD according to level."""
    fd = stdout if level not in ("ERROR",) else stderr

    if level == "INFO" and not verbose:
        return

    color_mapping: dict[MessageLevel, str] = {
        "ERROR": CLIColor.RED,
        "WARNING": CLIColor.YELLOW,
        "SUCCESS": CLIColor.GREEN,
    }
    print(
        f"{color_mapping.get(level, CLIColor.RESET)}[{level}] {text}{CLIColor.RESET}",
        file=fd,
    )



"""Assembler module to assemble programs in Gofra language into executables."""

from __future__ import annotations

import sys
from pathlib import Path
from platform import system as current_platform_system
from shutil import which
from subprocess import CalledProcessError, check_output
from typing import TYPE_CHECKING, Literal

from gofra.cli.output import cli_message
from gofra.codegen import generate_code_for_assembler
from gofra.codegen.backends.general import CODEGEN_ENTRY_POINT_SYMBOL
from gofra.codegen.get_backend import get_backend_for_target

from .exceptions import (
    NoToolkitForAssemblingError,
    UnsupportedBuilderOperatingSystemError,
)

if TYPE_CHECKING:
    from gofra.codegen.targets import TARGET_T
    from gofra.context import ProgramContext

type OUTPUT_FORMAT_T = Literal["library", "executable", "object", "assembly"]


def assemble_program(  # noqa: PLR0913
    context: ProgramContext,
    output: Path,
    output_format: OUTPUT_FORMAT_T,
    target: TARGET_T,
    *,
    build_cache_dir: Path,
    verbose: bool,
    additional_linker_flags: list[str],
    additional_assembler_flags: list[str],
    delete_build_cache_after_compilation: bool,
) -> None:
    """Convert given program into executable/library/etc using assembly and linker."""
    _validate_toolkit_installation()
    _prepare_build_cache_directory(build_cache_dir)

    assembly_filepath = _generate_assembly_file_with_codegen(
        context,
        target,
        output,
        build_cache_dir=build_cache_dir,
        verbose=verbose,
    )

    if output_format == "assembly":
        assembly_filepath.replace(output)
        return

    object_filepath = _assemble_object_file(
        target,
        assembly_filepath,
        output,
        additional_assembler_flags=additional_assembler_flags,
        build_cache_dir=build_cache_dir,
        verbose=verbose,
    )
    if output_format == "object":
        object_filepath.replace(output)
        if delete_build_cache_after_compilation:
            assembly_filepath.unlink()
        return

    assert output_format in ("executable", "library")
    _link_final_output(
        output,
        target,
        object_filepath,
        output_format=output_format,
        additional_linker_flags=additional_linker_flags,
        verbose=verbose,
    )

    if delete_build_cache_after_compilation:
        assembly_filepath.unlink()
        object_filepath.unlink()


def _prepare_build_cache_directory(build_cache_directory: Path) -> None:
    """Try to create and fill cache directory with required files."""
    if build_cache_directory.exists():
        return

    build_cache_directory.mkdir(exist_ok=False)

    with (build_cache_directory / ".gitignore").open("w") as f:
        f.write("# Do not include this newly generated build cache into git VCS\n")
        f.write("*\n")


def _link_final_output(  # noqa: PLR0913
    output: Path,
    target: TARGET_T,
    o_filepath: Path,
    output_format: Literal["executable", "library"],
    additional_linker_flags: list[str],
    *,
    verbose: bool,
) -> None:
    """Use linker to link object file into executable."""
    match current_platform_system():
        case "Darwin":
            assert target == "aarch64-darwin"

            system_sdk = Path(
                check_output(  # noqa: S603
                    ["/usr/bin/xcrun", "-sdk", "macosx", "--show-sdk-path"],
                    text=True,
                ).strip(),
            )
            target_linker_flags = [
                "-arch",
                "arm64",
                "-lSystem",
                "-syslibroot",
                str(system_sdk),
            ]

            if output_format == "library":
                target_linker_flags += ["-dylib"]
        case "Linux":
            assert target == "x86_64-linux"

            target_linker_flags = []
            assert output_format == "executable", (
                "Libraries on Linux is not implemented"
            )
        case _:
            raise UnsupportedBuilderOperatingSystemError

    linker_flags = [
        *target_linker_flags,
        *additional_linker_flags,
    ]

    if output_format == "executable":
        linker_flags += ["-e", CODEGEN_ENTRY_POINT_SYMBOL]

    command = ["/usr/bin/ld", "-o", str(output), str(o_filepath), *linker_flags]
    cli_message(
        level="INFO",
        text=f"Running linker command: `{' '.join(command)}`",
        verbose=verbose,
    )
    check_output(command)  # noqa: S603


def _assemble_object_file(  # noqa: PLR0913
    target: TARGET_T,
    asm_filepath: Path,
    output: Path,
    *,
    build_cache_dir: Path,
    additional_assembler_flags: list[str],
    verbose: bool,
) -> Path:
    """Call assembler to assemble given assembly file from codegen."""
    object_filepath = (build_cache_dir / output.name).with_suffix(".o")

    # Assembler is not crossplatform so we expect host has same architecture
    match current_platform_system():
        case "Darwin":
            if target != "aarch64-darwin":
                raise UnsupportedBuilderOperatingSystemError
            assembler_flags = ["-arch", "arm64"]
        case "Linux":
            if target != "x86_64-linux":
                raise UnsupportedBuilderOperatingSystemError
            assembler_flags = []
        case _:
            raise UnsupportedBuilderOperatingSystemError

    command = [
        "/usr/bin/as",
        "-o",
        str(object_filepath),
        str(asm_filepath),
        *assembler_flags,
        *additional_assembler_flags,
    ]
    cli_message(
        level="INFO",
        text=f"Running command: `{' '.join(command)}`",
        verbose=verbose,
    )
    try:
        check_output(command)  # noqa: S603
    except CalledProcessError as e:
        cli_message(
            "ERROR",
            "Failed to generate binary from output assembly, "
            f"error code: {e.returncode}",
        )
        sys.exit(1)

    return object_filepath


def _generate_assembly_file_with_codegen(
    context: ProgramContext,
    target: TARGET_T,
    output: Path,
    *,
    build_cache_dir: Path,
    verbose: bool,
) -> Path:
    """Call desired codegen backend for requested target and generate file contains assembly."""
    assembly_filepath = (build_cache_dir / output.name).with_suffix(".s")

    infered_backend = get_backend_for_target(target).__name__  # type: ignore  # noqa: PGH003
    cli_message(
        level="INFO",
        text=f"Generating assembly using codegen backend (Infered codegen for target `{target}` is `{infered_backend}`)...",
        verbose=verbose,
    )
    generate_code_for_assembler(assembly_filepath, context, target)
    return assembly_filepath


def _validate_toolkit_installation() -> None:
    """Validate that the host system has all requirements installed (linker/assembler)."""
    match current_platform_system():
        case "Darwin":
            required_toolkit = ("as", "ld", "xcrun")
        case "Linux":
            required_toolkit = ("as", "ld")
        case _:
            raise UnsupportedBuilderOperatingSystemError
    toolkit = {(tk, which(tk) is not None) for tk in required_toolkit}
    missing_toolkit = {tk for (tk, tk_is_installed) in toolkit if not tk_is_installed}
    if missing_toolkit:
        raise NoToolkitForAssemblingError(toolkit_required=missing_toolkit)



from gofra.codegen.targets import TARGET_T
from gofra.exceptions import GofraError


class CodegenUnsupportedBackendTargetPairError(GofraError):
    def __init__(
        self,
        *args: object,
        target: TARGET_T,
    ) -> None:
        super().__init__(*args)
        self.target = target

    def __repr__(self) -> str:
        return f"""Code generation failed

Unsupported target '{self.target}'!
Please read documentation to find available target pairs!
"""



from gofra.context import ProgramContext
from gofra.parser.functions.function import Function
from gofra.parser.operators import OperatorType

DCE_MAX_TRAVERSIONS = 128


def optimize_dead_code_elimination(
    program: ProgramContext,
) -> None:
    """Remove dead code from final operators result."""
    dce_remove_unused_functions(program)


def dce_remove_unused_functions(program: ProgramContext) -> None:
    """Apply DCE for functions so unused functions are removed."""
    for _ in range(DCE_MAX_TRAVERSIONS):
        unused_functions = [
            function
            for function in program.functions.values()
            if not _is_function_was_called(program, program.functions[function.name])
        ]
        if not unused_functions:
            return

        [program.functions.pop(function.name) for function in unused_functions]


def _is_function_was_called(program: ProgramContext, function: Function) -> bool:
    """Check is given function was called atleast once in whole program."""
    for possible_caller in [*program.functions.values(), program.entry_point]:
        for operator in possible_caller.source:
            if (
                operator.type == OperatorType.FUNCTION_CALL
                and operator.operand == function.name
            ):
                return True
    return False



from gofra.context import ProgramContext

from .strategies import (
    optimize_constant_folding,
    optimize_dead_code_elimination,
)


def optimize_program(program: ProgramContext) -> None:
    """Apply optimization strategies within given program context."""
    optimize_constant_folding(program)
    optimize_dead_code_elimination(program)



from collections import deque
from collections.abc import Generator, Iterable, MutableMapping
from pathlib import Path

from gofra.lexer import Token

from .macros import Macro


class PreprocessorState:
    """State of an preprocessor stage."""

    # Current file which is being processed
    path: Path

    # Lexical token streams from lexer or preprocessor itself
    # by default first is one from lexer and then it is extended by preprocessor to also consume next tokens from it until exhausted
    tokenizers: deque[Generator[Token]]

    # Remember which paths was included to not include them again.
    already_included_paths: list[Path]

    # Where to additionally search for paths
    include_search_paths: Iterable[Path]

    macros: MutableMapping[str, Macro]

    def __init__(
        self,
        path: Path,
        lexer: Generator[Token],
        include_search_paths: Iterable[Path],
    ) -> None:
        self.path = path.resolve(strict=True)

        self.include_search_paths = include_search_paths
        self.already_included_paths = [path]

        self.macros = {}

        self.tokenizers = deque((lexer,))
        self.tokenizer = self.iterate_tokenizers()

    def iterate_tokenizers(self) -> Generator[Token]:
        """Consume tokens from each tokenizers until all exhausted."""
        while self.tokenizers:
            # Consume token from current (last) tokenizer
            tokenizer = self.tokenizers[-1]
            token = next(tokenizer, None)

            if token:
                yield token
                continue

            # Current tokenizer exhausted and must be removed
            self.tokenizers.pop()



from enum import IntEnum, auto


class Intrinsic(IntEnum):
    INCREMENT = auto()
    DECREMENT = auto()

    PLUS = auto()
    MINUS = auto()
    MULTIPLY = auto()
    DIVIDE = auto()
    MODULUS = auto()

    EQUAL = auto()
    NOT_EQUAL = auto()
    LESS_THAN = auto()
    GREATER_THAN = auto()
    LESS_EQUAL_THAN = auto()
    GREATER_EQUAL_THAN = auto()

    MEMORY_LOAD = auto()
    MEMORY_STORE = auto()

    SYSCALL0 = auto()
    SYSCALL1 = auto()
    SYSCALL2 = auto()
    SYSCALL3 = auto()
    SYSCALL4 = auto()
    SYSCALL5 = auto()
    SYSCALL6 = auto()

    COPY = auto()
    DROP = auto()
    SWAP = auto()


WORD_TO_INTRINSIC = {
    "+": Intrinsic.PLUS,
    "-": Intrinsic.MINUS,
    "*": Intrinsic.MULTIPLY,
    "/": Intrinsic.DIVIDE,
    "==": Intrinsic.EQUAL,
    "!=": Intrinsic.NOT_EQUAL,
    "<": Intrinsic.LESS_THAN,
    ">": Intrinsic.GREATER_THAN,
    ">=": Intrinsic.LESS_EQUAL_THAN,
    "<=": Intrinsic.GREATER_EQUAL_THAN,
    "%": Intrinsic.MODULUS,
    "dec": Intrinsic.DECREMENT,
    "inc": Intrinsic.INCREMENT,
    "swap": Intrinsic.SWAP,
    "copy": Intrinsic.COPY,
    "drop": Intrinsic.DROP,
    "syscall0": Intrinsic.SYSCALL0,
    "syscall1": Intrinsic.SYSCALL1,
    "syscall2": Intrinsic.SYSCALL2,
    "syscall3": Intrinsic.SYSCALL3,
    "syscall4": Intrinsic.SYSCALL4,
    "syscall5": Intrinsic.SYSCALL5,
    "syscall6": Intrinsic.SYSCALL6,
    "?>": Intrinsic.MEMORY_LOAD,
    "!<": Intrinsic.MEMORY_STORE,
}



"""Assembly abstraction layer that hides declarative assembly OPs into functions that generates that for you."""

from __future__ import annotations

from typing import TYPE_CHECKING, assert_never

from .registers import (
    AARCH64_DOUBLE_WORD_BITS,
    AARCH64_GP_REGISTERS,
    AARCH64_HALF_WORD_BITS,
    AARCH64_MACOS_ABI_ARGUMENT_REGISTERS,
    AARCH64_MACOS_ABI_RETVAL_REGISTER,
    AARCH64_MACOS_SYSCALL_NUMBER_REGISTER,
    AARCH64_STACK_ALIGNMENT,
    AARCH64_STACK_ALINMENT_BIN,
)

if TYPE_CHECKING:
    from gofra.codegen.backends.aarch64_macos._context import AARCH64CodegenContext
    from gofra.codegen.backends.general import CODEGEN_GOFRA_ON_STACK_OPERATIONS


def drop_cells_from_stack(context: AARCH64CodegenContext, *, cells_count: int) -> None:
    """Drop stack cells from stack by shifting stack pointer (SP) by N bytes.

    Stack must be aligned so we use `AARCH64_STACK_ALIGNMENT`
    """
    assert cells_count > 0, "Tried to drop negative cells count from stack"
    stack_pointer_shift = AARCH64_STACK_ALIGNMENT * cells_count
    context.write(f"add SP, SP, #{stack_pointer_shift}")


def pop_cells_from_stack_into_registers(
    context: AARCH64CodegenContext,
    *registers: AARCH64_GP_REGISTERS,
) -> None:
    """Pop cells from stack and store into given registers.

    Each cell is 8 bytes, but stack pointer is always adjusted by `AARCH64_STACK_ALIGNMENT` bytes
    to preserve stack alignment.
    """
    assert registers, "Expected registers to store popped result into!"

    for register in registers:
        context.write(
            f"ldr {register}, [SP]",
            f"add SP, SP, #{AARCH64_STACK_ALIGNMENT}",
        )


def push_register_onto_stack(
    context: AARCH64CodegenContext,
    register: AARCH64_GP_REGISTERS,
) -> None:
    """Store given register onto stack under current stack pointer."""
    context.write(f"str {register}, [SP, -{AARCH64_STACK_ALIGNMENT}]!")


def store_integer_into_register(
    context: AARCH64CodegenContext,
    register: AARCH64_GP_REGISTERS,
    value: int,
) -> None:
    """Store given value into given register."""
    context.write(f"mov {register}, #{value}")


def push_integer_onto_stack(
    context: AARCH64CodegenContext,
    value: int,
) -> None:
    """Push given integer onto stack with auto shifting less-significant bytes.

    Value must be less than 16 bytes (18_446_744_073_709_551_615).
    Negative numbers is dissalowed.

    TODO(@kirillzhosul): Negative numbers IS dissalowed:
        Consider using signed two complement representation with sign bit (highest bit) set
    """
    assert value >= 0, "Tried to push negative integer onto stack!"
    assert value <= AARCH64_DOUBLE_WORD_BITS, (
        "Tried to push integer that exceeding 16 bytes (64 bits register)."
    )

    if value <= AARCH64_HALF_WORD_BITS:
        # We have small immediate value which we may just store without shifts
        context.write(f"mov X0, #{value}")
        push_register_onto_stack(context, register="X0")
        return

    preserve_bits = False
    for shift in range(0, 64, 16):
        chunk = hex((value >> shift) & AARCH64_HALF_WORD_BITS)
        if chunk == "0x0":
            # Zeroed chunk so we dont push it as register is zerod
            continue

        if not preserve_bits:
            # Store upper bits
            context.write(f"movz X0, #{chunk}, lsl #{shift}")
            preserve_bits = True
            continue

        # Store lower bits
        context.write(f"movk X0, #{chunk}, lsl #{shift}")

    push_register_onto_stack(context, register="X0")


def push_static_address_onto_stack(
    context: AARCH64CodegenContext,
    segment: str,
) -> None:
    """Push executable static memory addresss onto stack with page dereference."""
    context.write(
        f"adrp X0, {segment}@PAGE",
        f"add X0, X0, {segment}@PAGEOFF",
    )
    push_register_onto_stack(context, register="X0")


def evaluate_conditional_block_on_stack_with_jump(
    context: AARCH64CodegenContext,
    jump_over_label: str,
) -> None:
    """Evaluate conditional block by popping current value under SP againts zero.

    If condition is false (value on stack) then jump out that conditional block to `jump_over_label`
    """
    pop_cells_from_stack_into_registers(context, "X0")
    context.write(
        "cmp X0, #0",
        f"beq {jump_over_label}",
    )


def initialize_static_data_section(
    context: AARCH64CodegenContext,
    static_data_section: list[tuple[str, str | int]],
) -> None:
    """Initialize data section fields with given values.

    Section is an tuple (label, data)
    Data is an string (raw ASCII) or number (zeroed memory blob)
    """
    context.fd.write(".section __DATA,__data\n")
    context.fd.write(f".align {AARCH64_STACK_ALINMENT_BIN}\n")

    for name, data in static_data_section:
        if isinstance(data, str):
            context.fd.write(f'{name}: .asciz "{data}"\n')
            continue
        context.fd.write(f"{name}: .space {data}\n")


def ipc_syscall_macos(
    context: AARCH64CodegenContext,
    *,
    arguments_count: int,
    store_retval_onto_stack: bool,
    injected_args: list[int | None] | None,
) -> None:
    """Call system (syscall) via supervisor call and apply IPC ABI convention to arguments."""
    assert not injected_args or len(injected_args) == arguments_count + 1

    if not injected_args:
        injected_args = [None for _ in range(arguments_count + 1)]

    registers_to_load = (
        AARCH64_MACOS_SYSCALL_NUMBER_REGISTER,
        *AARCH64_MACOS_ABI_ARGUMENT_REGISTERS[:arguments_count][::-1],
    )

    for injected_argument, register in zip(injected_args, registers_to_load):
        if injected_argument is not None:
            # Register injected and infered from stack
            store_integer_into_register(
                context,
                register=register,
                value=injected_argument,
            )
            continue
        pop_cells_from_stack_into_registers(context, register)

    # Supervisor call (syscall)
    context.write("svc #0")

    if store_retval_onto_stack:
        # Mostly related to optimizations above if we dont want to store result
        push_register_onto_stack(context, AARCH64_MACOS_ABI_RETVAL_REGISTER)


def perform_operation_onto_stack(
    context: AARCH64CodegenContext,
    operation: CODEGEN_GOFRA_ON_STACK_OPERATIONS,
) -> None:
    """Perform *math* operation onto stack (pop arguments and push back result)."""
    is_unary = operation in ("++", "--")
    registers = ("X0",) if is_unary else ("X0", "X1")
    pop_cells_from_stack_into_registers(context, *registers)

    match operation:
        case "+":
            context.write("add X0, X1, X0")
        case "-":
            context.write("sub X0, X1, X0")
        case "*":
            context.write("mul X0, X1, X0")
        case "//":
            context.write("sdiv X0, X1, X0")
        case "%":
            context.write(
                "udiv X2, X1, X0",
                "mul X2, X2, X0",
                "sub X0, X1, X2",
            )
        case "++":
            context.write("add X0, X0, #1")
        case "--":
            context.write("sub X0, X0, #1")
        case "!=" | ">=" | "<=" | "<" | ">" | "==":
            logic_op = {
                "!=": "ne",
                ">=": "ge",
                "<=": "le",
                "<": "lt",
                ">": "gt",
                "==": "eq",
            }
            context.write(
                "cmp X1, X0",
                f"cset X0, {logic_op[operation]}",
            )
        case _:
            assert_never()
    push_register_onto_stack(context, "X0")


def load_memory_from_stack_arguments(context: AARCH64CodegenContext) -> None:
    """Load memory as value using arguments from stack."""
    pop_cells_from_stack_into_registers(context, "X0")
    context.write("ldr X0, [X0]")
    push_register_onto_stack(context, "X0")


def store_into_memory_from_stack_arguments(context: AARCH64CodegenContext) -> None:
    """Store value from into memory pointer, pointer and value acquired from stack."""
    pop_cells_from_stack_into_registers(context, "X0", "X1")
    context.write("str X0, [X1]")


def call_function_block(
    context: AARCH64CodegenContext,
    function_name: str,
    *,
    abi_ffi_push_retval_onto_stack: bool,
    abi_ffi_arguments_count: int,
) -> None:
    """Call an function with preparing required fields (like stack-pointer).

    Also allows to call ABI/FFI function with providing arguments and return value (retval) via:
    `abi_ffi_push_retval_onto_stack` / `abi_ffi_arguments_count`
    """
    assert abi_ffi_arguments_count >= 0, "FFI arguments count cannot go negative"
    if abi_ffi_arguments_count:
        arguments = abi_ffi_arguments_count
        registers = AARCH64_MACOS_ABI_ARGUMENT_REGISTERS[:arguments][::-1]
        pop_cells_from_stack_into_registers(context, *registers)

    context.write(f"bl {function_name}")

    if abi_ffi_push_retval_onto_stack:
        push_register_onto_stack(context, AARCH64_MACOS_ABI_RETVAL_REGISTER)


def function_begin_with_prologue(
    context: AARCH64CodegenContext,
    *,
    function_name: str,
    as_global_linker_symbol: bool,
) -> None:
    """Begin an function symbol with prologue with preparing required (like stack-pointer)."""
    if as_global_linker_symbol:
        context.fd.write(f".global {function_name}\n")
    context.fd.write(f".align {AARCH64_STACK_ALINMENT_BIN}\n")
    context.fd.write(f"{function_name}:\n")


def function_end_with_epilogue(context: AARCH64CodegenContext) -> None:
    """Functions epilogue at the end. Restores required fields (like stack-pointer)."""
    context.write("ret")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from gofra.lexer.keywords import Keyword
from gofra.lexer.lexer import tokenize_file
from gofra.lexer.tokens import Token, TokenType

from .exceptions import (
    PreprocessorIncludeCurrentFileError,
    PreprocessorIncludeFileNotFoundError,
    PreprocessorIncludeNonStringNameError,
    PreprocessorIncludeNoPathError,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from gofra.preprocessor._state import PreprocessorState


def resolve_include_from_token_into_state(
    include_token: Token,
    state: PreprocessorState,
) -> None:
    """Consume and resolve include construction into real include at preprocessor side, ommiting include if already included."""
    requested_include_path = _consume_include_raw_path_from_token(include_token, state)
    if requested_include_path.resolve(strict=False) == state.path:
        raise PreprocessorIncludeCurrentFileError(include_path_token=include_token)

    include_path = _try_resolve_and_find_real_include_path(
        requested_include_path,
        current_path=include_token.location.filepath,
        search_paths=state.include_search_paths,
    )

    if include_path is None:
        raise PreprocessorIncludeFileNotFoundError(
            include_token=include_token,
            include_path=requested_include_path,
        )

    if include_path not in state.already_included_paths:
        state.already_included_paths.append(include_path)
        state.tokenizers.append(tokenize_file(include_path))


def _consume_include_raw_path_from_token(
    include_token: Token,
    state: PreprocessorState,
) -> Path:
    """Consume include path from `include` construction."""
    assert include_token.type == TokenType.KEYWORD
    assert include_token.value == Keyword.PP_INCLUDE

    include_path_token = next(state.tokenizer, None)
    if not include_path_token:
        raise PreprocessorIncludeNoPathError(include_token=include_token)
    if include_path_token.type != TokenType.STRING:
        raise PreprocessorIncludeNonStringNameError(
            include_path_token=include_path_token,
        )

    include_path_raw = include_path_token.value
    assert isinstance(include_path_raw, str)
    return Path(include_path_raw)


def _try_resolve_and_find_real_include_path(
    path: Path,
    current_path: Path,
    search_paths: Iterable[Path],
) -> Path | None:
    """Resolve real import path and try to search for possible location of include (include directories system)."""
    for search_path in (Path("./"), current_path.parent, *search_paths):
        if (probable_path := search_path.joinpath(path)).exists(follow_symlinks=True):
            if probable_path.is_file():
                # We found an straighforward file reference
                return probable_path

            # Non-existant file here or directory reference.
            if not probable_path.is_dir():
                continue

            probable_package = Path(probable_path / probable_path.name).with_suffix(
                ".gof",
            )
            if probable_package.exists():
                return probable_package
    return None



"""Optimization strategies applied to the program to optimize it."""

from .constant_folding import optimize_constant_folding
from .dead_code_elimination import optimize_dead_code_elimination

__all__ = ["optimize_constant_folding", "optimize_dead_code_elimination"]



from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import IO


@dataclass(frozen=True)
class AARCH64CodegenContext:
    """General context for emitting code from IR.

    Probably this is weird and bad way but OK for now.
    @kirillzhosul: Refactor at some point
    """

    fd: IO[str]
    strings: MutableMapping[str, str] = field()

    def write(self, *lines: str) -> int:
        return self.fd.write("\t" + "\n\t".join(lines) + "\n")

    def load_string(self, string: str) -> str:
        string_key = "str%d" % len(self.strings)
        self.strings[string_key] = string
        return string_key



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.lexer.keywords import WORD_TO_KEYWORD, Keyword
from gofra.lexer.tokens import TokenType
from gofra.parser.intrinsics import WORD_TO_INTRINSIC

from .container import Macro
from .exceptions import (
    PreprocessorMacroDefinesMacroError,
    PreprocessorMacroNonWordNameError,
    PreprocessorMacroRedefinedError,
    PreprocessorMacroRedefinesLanguageDefinitionError,
    PreprocessorNoMacroNameError,
    PreprocessorUnclosedMacroError,
)

if TYPE_CHECKING:
    from gofra.lexer import Token
    from gofra.preprocessor._state import PreprocessorState

PROHIBITED_MACRO_NAMES = WORD_TO_INTRINSIC.keys() | WORD_TO_KEYWORD.keys()


def define_macro_block_from_token(token: Token, state: PreprocessorState) -> None:
    macro_name = _consume_macro_name_from_token(token, state)
    macro = _define_empty_macro(state, token, macro_name)
    _consume_tokenizer_from_state_into_macro(macro, state)


def try_resolve_macro_reference_from_token(
    token: Token,
    state: PreprocessorState,
) -> bool:
    """Try to search for defined macro and resolve it if possible."""
    assert token.type == TokenType.WORD
    assert isinstance(token.value, str)

    macro = state.macros.get(token.value)
    if not macro:
        return False

    state.tokenizers.append(macro.as_tokenizer())
    return True


def _consume_macro_name_from_token(token: Token, state: PreprocessorState) -> str:
    """Consume macro name from `macro` construction."""
    macro_name_token = next(state.tokenizer, None)
    if not macro_name_token:
        raise PreprocessorNoMacroNameError(macro_token=token)
    if macro_name_token.type != TokenType.WORD:
        raise PreprocessorMacroNonWordNameError(macro_name_token=macro_name_token)

    macro_name = macro_name_token.text
    if macro_name in state.macros:
        raise PreprocessorMacroRedefinedError(
            redefine_macro_name_token=macro_name_token,
            original_macro_location=state.macros[macro_name].token.location,
        )
    if macro_name in PROHIBITED_MACRO_NAMES:
        raise PreprocessorMacroRedefinesLanguageDefinitionError(
            macro_token=macro_name_token,
            macro_name=macro_name,
        )
    return macro_name


def _define_empty_macro(
    state: PreprocessorState,
    token: Token,
    name: str,
) -> Macro:
    """Create empty macro that start at given token to fill it preprocessed tokens."""
    macro = Macro(token=token, tokens=[], name=name)
    state.macros[name] = macro
    return macro


def _consume_tokenizer_from_state_into_macro(
    macro: Macro,
    state: PreprocessorState,
) -> None:
    """Consume current tokenizer into macro block."""
    opened_context_blocks = 0
    macro_was_closed = False

    while token := next(state.tokenizer, None):
        if token.type == TokenType.KEYWORD:
            if token.value == Keyword.PP_MACRO:
                raise PreprocessorMacroDefinesMacroError(
                    macro_token=macro.token,
                    macro_name=macro.name,
                )
            if token.value == Keyword.END:
                if opened_context_blocks <= 0:
                    macro_was_closed = True
                    break
                opened_context_blocks -= 1
            if token.value in (Keyword.IF, Keyword.DO, Keyword.FUNCTION):
                opened_context_blocks += 1
        macro.tokens.append(token)
    if not macro_was_closed:
        raise PreprocessorUnclosedMacroError(
            macro_token=macro.token,
            macro_name=macro.name,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.exceptions import GofraError
from gofra.typecheck.types import WORD_TO_GOFRA_TYPE

if TYPE_CHECKING:
    from gofra.lexer.tokens import Token


class ParserFunctionIsBothInlineAndExternalError(GofraError):
    def __init__(self, *args: object, modifier_token: Token) -> None:
        super().__init__(*args)
        self.modifier_token = modifier_token

    def __repr__(self) -> str:
        return f"""Tried to define function that is both `external` and `inline`.
Found at: {self.modifier_token.location}

External function cannot have an body and defined outside, 
which leads to behavior for `extern inline` function unknown and prohibited"""


class ParserFunctionModifierReappliedError(GofraError):
    def __init__(self, *args: object, modifier_token: Token) -> None:
        super().__init__(*args)
        self.modifier_token = modifier_token

    def __repr__(self) -> str:
        return f"""Function modifier reapplied.
Found at: {self.modifier_token.location}

There is no need to reapply same function modifier."""


class ParserExpectedFunctionAfterFunctionModifiersError(GofraError):
    def __init__(self, *args: object, modifier_token: Token) -> None:
        super().__init__(*args)
        self.modifier_token = modifier_token

    def __repr__(self) -> str:
        return f"""Expected function keyword after function modifiers (`inline`, `extern`).
But last token at {self.modifier_token.location} is not an function keyword!

Function modifiers must be followed by function keyword."""


class ParserExpectedFunctionReturnTypeError(GofraError):
    def __init__(self, *args: object, definition_token: Token) -> None:
        super().__init__(*args)
        self.definition_token = definition_token

    def __repr__(self) -> str:
        return f"""Expected function return type at {self.definition_token.location}.

Do you have unfinished function definition?"""


class ParserFunctionNoNameError(GofraError):
    def __init__(self, *args: object, token: Token) -> None:
        super().__init__(*args)
        self.token = token

    def __repr__(self) -> str:
        return f"""Expected function name with signature at {self.token.location}!

Do you have unfinished function definition?"""


class ParserExpectedFunctionKeywordError(GofraError):
    def __init__(self, *args: object, token: Token) -> None:
        super().__init__(*args)
        self.token = token

    def __repr__(self) -> str:
        return f"""Expected function keyword at {self.token.location}!

But got something else.
Do you have unfinished function definition?"""


class ParserFunctionInvalidTypeError(GofraError):
    def __init__(self, *args: object, type_token: Token, requested_type: str) -> None:
        super().__init__(*args)
        self.type_token = type_token
        self.requested_type = requested_type

    def __repr__(self) -> str:
        return f"""Unknown type `{self.requested_type}` used in function definition at {self.type_token.location}.

Expected one of known types: {", ".join(WORD_TO_GOFRA_TYPE.keys())}"""



from __future__ import annotations

from typing import TYPE_CHECKING

from gofra.consts import GOFRA_ENTRY_POINT
from gofra.exceptions import GofraError

if TYPE_CHECKING:
    from collections.abc import Iterable

    from gofra.lexer.tokens import Token, TokenLocation
    from gofra.parser.functions.function import FunctionTypeContract


class ParserDirtyNonPreprocessedTokenError(GofraError):
    def __init__(self, *args: object, token: Token) -> None:
        super().__init__(*args)
        self.token = token

    def __repr__(
        self,
    ) -> str:
        return f"""Got dirty non preprocessed token from tokenizer at parser stage.

`{self.token.text}` at {self.token.location}

Probably this is not an language user fault.
Did you forgot to call preprocessor stage to resolve these dirty tokens?"""


class ParserExhaustiveContextStackError(GofraError):
    def __repr__(self) -> str:
        return """Exhaustive context stack!
All context blocks should be folded at the end

Do you have not closed block?"""


class ParserUnfinishedWhileDoBlockError(GofraError):
    def __init__(self, *args: object, token: Token) -> None:
        super().__init__(*args)
        self.token = token

    def __repr__(self) -> str:
        return f"""Unclosed 'while ... do' block at {self.token.location}!
Expected there will be 'end' block after 'do' and 'do' after 'while'.

Did you forgot to open/close 'do' block?"""


class ParserUnfinishedIfBlockError(GofraError):
    def __init__(self, *args: object, if_token: Token) -> None:
        super().__init__(*args)
        self.if_token = if_token

    def __repr__(self) -> str:
        return f"""Unclosed 'if' block at {self.if_token.location}!
Expected there will be 'end' block after 'if'.

Did you forgot to close 'if' block?"""


class ParserUnknownWordError(GofraError):
    def __init__(
        self,
        *args: object,
        word_token: Token,
        macro_names: Iterable[str],
        best_match: str | None = None,
    ) -> None:
        super().__init__(*args)
        self.word_token = word_token
        self.macro_names = macro_names
        self.best_match = best_match

    def __repr__(self) -> str:
        return f"""Encountered an unknown name '{self.word_token.text}' at {self.word_token.location}!
Expected intrinsic or macro name.

Available macro names: {", ".join(self.macro_names) or "..."}""" + (
            f"\nDid you mean '{self.best_match}'?" if self.best_match else ""
        )


class ParserNoWhileBeforeDoError(GofraError):
    def __init__(self, *args: object, do_token: Token) -> None:
        super().__init__(*args)
        self.do_token = do_token

    def __repr__(self) -> str:
        return f"""No 'while' before 'do' at {self.do_token.location}!
Expected there will be 'while' block before 'do'.

Did you forgot to add starting block?"""


class ParserEndAfterWhileError(GofraError):
    def __init__(self, *args: object, end_token: Token) -> None:
        super().__init__(*args)
        self.end_token = end_token

    def __repr__(self) -> str:
        return f"""'end' after 'while' block  at {self.end_token.location}!
Expected there will be 'do' block after 'while'.

Did you forgot to add 'do' block?"""


class ParserEndWithoutContextError(GofraError):
    def __init__(self, *args: object, end_token: Token) -> None:
        super().__init__(*args)
        self.end_token = end_token

    def __repr__(self) -> str:
        return f"""'end' used without context at {self.end_token.location}!
Expected there will be context block before 'end'.

Did you forgot to add context block?"""


class ParserNoWhileConditionOperatorsError(GofraError):
    def __init__(self, *args: object, while_token: Token) -> None:
        super().__init__(*args)
        self.while_token = while_token

    def __repr__(self) -> str:
        return f"""'while ... do' used without condition inside at {self.while_token.location}!
Expected there will be at least something in condition before 'do'.
This will lead to infinite loop, which may cause undefined behavior.
Please use 'while true do .. end' if this is your intense.

Did you forgot to add condition?"""


class ExternNoFunctionNameError(GofraError):
    def __init__(self, *args: object, macro_token: Token) -> None:
        super().__init__(*args)
        self.macro_token = macro_token

    def __repr__(self) -> str:
        return f"""No 'extern' function name specified at {self.macro_token.location}!
Extern functions should have name after 'extern' keyword

Do you have unfinished extern function definition?"""


class ParserExternNonWordNameError(GofraError):
    def __init__(self, *args: object, function_name_token: Token) -> None:
        super().__init__(*args)
        self.function_name_token = function_name_token

    def __repr__(self) -> str:
        return f"""Non word name for extern function name at {self.function_name_token.location}!
Extern function signature should have name as word after 'extern' keyword but got '{self.function_name_token.type.name}'!"""


class ParserExternRedefinesMacroError(GofraError):
    def __init__(
        self,
        *args: object,
        redefine_extern_function_name_token: Token,
        original_macro_location: TokenLocation,
        original_macro_name: str,
    ) -> None:
        super().__init__(*args)
        self.redefine_extern_function_name_token = redefine_extern_function_name_token
        self.original_macro_location = original_macro_location
        self.original_macro_name = original_macro_name

    def __repr__(self) -> str:
        return f"""Redefinition of an macro '{self.original_macro_name}' at {self.redefine_extern_function_name_token.location} inside extern function!
Original definition found at {self.original_macro_location}!
Extern cannot redefine macro names!"""


class ParserExternRedefinesLanguageDefinitionError(GofraError):
    def __init__(
        self,
        *args: object,
        extern_token: Token,
        extern_function_name: str,
    ) -> None:
        super().__init__(*args)
        self.extern_token = extern_token
        self.extern_function_name = extern_function_name

    def __repr__(self) -> str:
        return (
            f"Extern function '{self.extern_function_name}' at {self.extern_token.location}"
            " tries to redefine language definition!"
        )


class ParserEmptyIfBodyError(GofraError):
    def __init__(self, *args: object, if_token: Token) -> None:
        super().__init__(*args)
        self.if_token = if_token

    def __repr__(self) -> str:
        return f"""If condition at {self.if_token.location} has no body!
If condition should always have body as otherwise this has no effect!"""


class ParserNoEntryFunctionError(GofraError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __repr__(self) -> str:
        return f"""Expected entry point function '{GOFRA_ENTRY_POINT}' but it does not exists!"""


class ParserEntryPointFunctionTypeContractInError(GofraError):
    def __init__(self, *args: object, type_contract_in: FunctionTypeContract) -> None:
        super().__init__(*args)
        self.type_contract_in = type_contract_in

    def __repr__(self) -> str:
        return f"""Entry point function '{GOFRA_ENTRY_POINT}' violates type contract!

Entry point function cannot have type contract in!
But currently it have type contract in: {self.type_contract_in}
"""


class ParserEntryPointFunctionTypeContractOutError(GofraError):
    def __init__(self, *args: object, type_contract_out: FunctionTypeContract) -> None:
        super().__init__(*args)
        self.type_contract_out = type_contract_out

    def __repr__(self) -> str:
        return f"""Entry point function '{GOFRA_ENTRY_POINT}' violates type contract!

Entry point function cannot have type contract out!
But currently it have type contract out: {self.type_contract_out}
"""


class ParserEntryPointFunctionModifiersError(GofraError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __repr__(self) -> str:
        return f"""Entry point function '{GOFRA_ENTRY_POINT}' violates modifiers signature!

Entry point function cannot be external or inlined!
"""


class ParserTopLevelError(GofraError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __repr__(self) -> str:
        return """Expected

Entry point function cannot be external or inlined!
"""



from collections.abc import Generator
from pathlib import Path
from typing import IO

from .exceptions import IOFileDoesNotExistsError, IOFileNotAnFileError


def open_source_file_line_stream(path: Path) -> Generator[str]:
    """Open IO for an text file with Gofra source code and yield each line from it until it ends.

    :raises IOFileDoesNotExistsError: File does not exists
    :raises IOFileNotAnFileError: File is not an file
    """
    with open_source_file(path) as io:
        while line := io.readline(-1):
            yield line


def open_source_file(path: Path) -> IO[str]:
    """Open IO for an text file with Gofra source code.

    :raises IOFileDoesNotExistsError: File does not exists
    :raises IOFileNotAnFileError: File is not an file
    """
    if not path.exists(follow_symlinks=True):
        raise IOFileDoesNotExistsError(path=path)

    if not path.is_file():
        raise IOFileNotAnFileError(path=path)

    return path.open(
        mode="rt",
        buffering=-1,
        encoding="utf-8",
        errors="strict",
        newline=None,
    )



from enum import IntEnum, auto


class GofraType(IntEnum):
    """Types that are defined within language so we can check type safety."""

    # Number which size is infered from target backend (mostly 32/64 bits architectures)
    INTEGER = auto()

    # Infer integer as pointer for pointer arithmetics typecheck and FFI
    POINTER = auto()

    # Infer integer as boolean or result from comparsion
    BOOLEAN = auto()

    # Used to specify that function returns nothing
    VOID = auto()

    # Typechecker only!
    ANY = auto()

    def __repr__(self) -> str:
        return self.name


WORD_TO_GOFRA_TYPE = {
    "int": GofraType.INTEGER,
    "ptr": GofraType.POINTER,
    "bool": GofraType.BOOLEAN,
    "void": GofraType.VOID,
}



"""Entry point validator.

Probably should not be here (in parser)
"""

from gofra.consts import GOFRA_ENTRY_POINT
from gofra.parser.exceptions import (
    ParserEntryPointFunctionModifiersError,
    ParserEntryPointFunctionTypeContractInError,
    ParserEntryPointFunctionTypeContractOutError,
    ParserNoEntryFunctionError,
)
from gofra.parser.functions.function import Function

from ._context import ParserContext


def validate_and_pop_entry_point(context: ParserContext) -> Function:
    """Validate program entry, check its existance and type contracts."""
    if GOFRA_ENTRY_POINT not in context.functions:
        raise ParserNoEntryFunctionError

    entry_point = context.functions.pop(GOFRA_ENTRY_POINT)
    if entry_point.is_externally_defined or entry_point.emit_inline_body:
        raise ParserEntryPointFunctionModifiersError

    if entry_point.type_contract_out:
        raise ParserEntryPointFunctionTypeContractOutError(
            type_contract_out=entry_point.type_contract_out,
        )

    if entry_point.type_contract_out:
        raise ParserEntryPointFunctionTypeContractInError(
            type_contract_in=entry_point.type_contract_in,
        )
    return entry_point



from gofra.lexer.tokens import Token, TokenLocation
from gofra.preprocessor.exceptions import PreprocessorError


class PreprocessorNoMacroNameError(PreprocessorError):
    def __init__(self, *args: object, macro_token: Token) -> None:
        super().__init__(*args)
        self.macro_token = macro_token

    def __repr__(self) -> str:
        return f"""No 'macro' name specified at {self.macro_token.location}!
Macros should have name after 'macro' keyword as word

Do you have unfinished macro definition?"""


class PreprocessorMacroNonWordNameError(PreprocessorError):
    def __init__(self, *args: object, macro_name_token: Token) -> None:
        super().__init__(*args)
        self.macro_name_token = macro_name_token

    def __repr__(self) -> str:
        return f"""Non word name for macro at {self.macro_name_token.location}!

Macros should have name as word after 'macro' keyword but got '{self.macro_name_token.type.name}'!"""


class PreprocessorMacroRedefinedError(PreprocessorError):
    def __init__(
        self,
        *args: object,
        redefine_macro_name_token: Token,
        original_macro_location: TokenLocation,
    ) -> None:
        super().__init__(*args)
        self.redefine_macro_name_token = redefine_macro_name_token
        self.original_macro_location = original_macro_location

    def __repr__(self) -> str:
        return f"""Redefinition of an macro '{self.redefine_macro_name_token.text}' at {self.redefine_macro_name_token.location}

Original definition found at {self.original_macro_location}.

Only single definition allowed for macros."""


class PreprocessorMacroRedefinesLanguageDefinitionError(PreprocessorError):
    def __init__(self, *args: object, macro_token: Token, macro_name: str) -> None:
        super().__init__(*args)
        self.macro_token = macro_token
        self.macro_name = macro_name

    def __repr__(self) -> str:
        return f"""Macro '{self.macro_name}' at {self.macro_token.location} tries to redefine language definition!"""


class PreprocessorUnclosedMacroError(PreprocessorError):
    def __init__(self, *args: object, macro_token: Token, macro_name: str) -> None:
        super().__init__(*args)
        self.macro_token = macro_token
        self.macro_name = macro_name

    def __repr__(self) -> str:
        return f"""Unclosed macro '{self.macro_name}' at {self.macro_token.location}!

Macro definition should have 'end' to close block.

Did you forgot to close macro definition?"""


class PreprocessorMacroDefinesMacroError(PreprocessorError):
    def __init__(self, *args: object, macro_token: Token, macro_name: str) -> None:
        super().__init__(*args)
        self.macro_token = macro_token
        self.macro_name = macro_name

    def __repr__(self) -> str:
        return f"""Macro '{self.macro_name}' at {self.macro_token.location} defines another macro inside!

Macro cannot define another macro inside their bodies."""



"""Parser package that used to parse source tokens into operators."""

from .operators import Operator, OperatorType
from .parser import parse_file

__all__ = ["Operator", "OperatorType", "parse_file"]



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum, auto
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(frozen=True)
class TokenLocation:
    """Location of any token within source code file."""

    filepath: Path
    line_number: int
    col_number: int

    def __repr__(self) -> str:
        return f"'{self.filepath.name}:{self.line_number + 1}:{self.col_number + 1}'"


class TokenType(IntEnum):
    """Type of the lexical token.

    These types is a bit weird due to `convention` but this is a some sort of an legacy code.
    (e.g character/strings/integers must be an literal and parsed at PARSER side)
    https://en.wikipedia.org/wiki/Lexical_analysis
    """

    INTEGER = auto()

    CHARACTER = auto()
    STRING = auto()

    WORD = auto()
    KEYWORD = auto()


@dataclass(frozen=True)
class Token:
    """Lexical token obtained by lexer."""

    type: TokenType

    # Real text of an token within source code
    text: str

    # `pre-parsed` value (e.g numbers are numbers, string are unescaped)
    # This is actually may not be here (inside parser), but due to now this will be as-is
    value: int | str

    # Location within file
    location: TokenLocation



"""Core AARCH64 MacOS codegen."""

from __future__ import annotations

from typing import IO, TYPE_CHECKING, assert_never

from gofra.codegen.backends.general import (
    CODEGEN_ENTRY_POINT_SYMBOL,
    CODEGEN_GOFRA_CONTEXT_LABEL,
    CODEGEN_INTRINSIC_TO_ASSEMBLY_OPS,
)
from gofra.consts import GOFRA_ENTRY_POINT
from gofra.parser.functions.function import Function
from gofra.parser.intrinsics import Intrinsic
from gofra.parser.operators import Operator, OperatorType

from ._context import AMD64CodegenContext
from .assembly import (
    call_function_block,
    drop_cells_from_stack,
    evaluate_conditional_block_on_stack_with_jump,
    function_begin_with_prologue,
    function_end_with_epilogue,
    initialize_static_data_section,
    ipc_syscall_linux,
    load_memory_from_stack_arguments,
    perform_operation_onto_stack,
    pop_cells_from_stack_into_registers,
    push_integer_onto_stack,
    push_register_onto_stack,
    push_static_address_onto_stack,
    store_into_memory_from_stack_arguments,
)
from .registers import (
    AMD64_LINUX_EPILOGUE_EXIT_CODE,
    AMD64_LINUX_EPILOGUE_EXIT_SYSCALL_NUMBER,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from gofra.context import ProgramContext


def generate_amd64_linux_backend(
    fd: IO[str],
    program: ProgramContext,
) -> None:
    """AMD64 Linux code generation backend."""
    context = AMD64CodegenContext(fd=fd, strings={})

    context.write(".att_syntax noprefix")
    amd64_linux_executable_functions(context, program)
    amd64_linux_program_entry_point(context)
    amd64_linux_data_section(context, program)


def amd64_linux_instruction_set(
    context: AMD64CodegenContext,
    operators: Sequence[Operator],
    program: ProgramContext,
    owner_function_name: str,
) -> None:
    """Write executable instructions from given operators."""
    for idx, operator in enumerate(operators):
        amd64_linux_operator_instructions(
            context,
            operator,
            program,
            idx,
            owner_function_name,
        )
        if operator.type == OperatorType.FUNCTION_RETURN:
            break


def amd64_linux_operator_instructions(
    context: AMD64CodegenContext,
    operator: Operator,
    program: ProgramContext,
    idx: int,
    owner_function_name: str,
) -> None:
    match operator.type:
        case OperatorType.INTRINSIC:
            amd64_linux_intrinsic_instructions(context, operator)
        case OperatorType.PUSH_MEMORY_POINTER:
            assert isinstance(operator.operand, str)
            push_static_address_onto_stack(context, operator.operand)
        case OperatorType.PUSH_INTEGER:
            assert isinstance(operator.operand, int)
            push_integer_onto_stack(context, operator.operand)
        case OperatorType.DO | OperatorType.IF:
            assert isinstance(operator.jumps_to_operator_idx, int)
            label = CODEGEN_GOFRA_CONTEXT_LABEL % (
                owner_function_name,
                operator.jumps_to_operator_idx,
            )
            evaluate_conditional_block_on_stack_with_jump(context, label)
        case OperatorType.END | OperatorType.WHILE:
            # This also should be refactored into `assembly` layer
            label = CODEGEN_GOFRA_CONTEXT_LABEL % (owner_function_name, idx)
            if isinstance(operator.jumps_to_operator_idx, int):
                label_to = CODEGEN_GOFRA_CONTEXT_LABEL % (
                    owner_function_name,
                    operator.jumps_to_operator_idx,
                )
                context.write(f"jmp {label_to}")
            context.fd.write(f"{label}:\n")
        case OperatorType.PUSH_STRING:
            assert isinstance(operator.operand, str)
            push_static_address_onto_stack(
                context,
                segment=context.load_string(operator.token.text[1:-1]),
            )
            push_integer_onto_stack(context, value=len(operator.operand))
        case OperatorType.FUNCTION_CALL:
            assert isinstance(operator.operand, str)

            function = program.functions[operator.operand]
            call_function_block(
                context,
                function_name=function.name,
                abi_ffi_push_retval_onto_stack=function.abi_ffi_push_retval_onto_stack(),
                abi_ffi_arguments_count=len(function.type_contract_in),
            )
        case OperatorType.FUNCTION_RETURN:
            function_end_with_epilogue(context)
        case _:
            assert_never(operator.type)


def amd64_linux_intrinsic_instructions(
    context: AMD64CodegenContext,
    operator: Operator,
) -> None:
    """Write executable body for intrinsic operation."""
    assert isinstance(operator.operand, Intrinsic)
    assert operator.type == OperatorType.INTRINSIC
    match operator.operand:
        case Intrinsic.DROP:
            drop_cells_from_stack(context, cells_count=1)
        case Intrinsic.COPY:
            pop_cells_from_stack_into_registers(context, "rax")
            push_register_onto_stack(context, "rax")
            push_register_onto_stack(context, "rax")
        case Intrinsic.SWAP:
            pop_cells_from_stack_into_registers(context, "rax", "rbx")
            push_register_onto_stack(context, "rax")
            push_register_onto_stack(context, "rbx")
        case (
            Intrinsic.PLUS
            | Intrinsic.MINUS
            | Intrinsic.MULTIPLY
            | Intrinsic.DIVIDE
            | Intrinsic.MODULUS
            | Intrinsic.INCREMENT
            | Intrinsic.DECREMENT
            | Intrinsic.NOT_EQUAL
            | Intrinsic.GREATER_EQUAL_THAN
            | Intrinsic.LESS_EQUAL_THAN
            | Intrinsic.LESS_THAN
            | Intrinsic.GREATER_THAN
            | Intrinsic.EQUAL
        ):
            perform_operation_onto_stack(
                context,
                operation=CODEGEN_INTRINSIC_TO_ASSEMBLY_OPS[operator.operand],
            )
        case (
            Intrinsic.SYSCALL0
            | Intrinsic.SYSCALL1
            | Intrinsic.SYSCALL2
            | Intrinsic.SYSCALL3
            | Intrinsic.SYSCALL4
            | Intrinsic.SYSCALL5
            | Intrinsic.SYSCALL6
        ):
            assert operator.syscall_optimization_injected_args is None, "TODO: Optimize"
            ipc_syscall_linux(
                context,
                arguments_count=operator.get_syscall_arguments_count() - 1,
                store_retval_onto_stack=not operator.syscall_optimization_omit_result,
                injected_args=[],
            )
        case Intrinsic.MEMORY_LOAD:
            load_memory_from_stack_arguments(context)
        case Intrinsic.MEMORY_STORE:
            store_into_memory_from_stack_arguments(context)
        case _:
            assert_never(operator.operand)


def amd64_linux_executable_functions(
    context: AMD64CodegenContext,
    program: ProgramContext,
) -> None:
    """Define all executable functions inside final executable with their executable body respectuflly.

    Provides an prolog and epilogue.
    """
    # Define only function that contains anything to execute
    functions = filter(
        Function.has_executable_body,
        [*program.functions.values(), program.entry_point],
    )
    for function in functions:
        assert not function.is_global_linker_symbol or (
            not function.type_contract_in and not function.type_contract_out
        ), "Codegen does not supports global linker symbols that has type contracts"
        function_begin_with_prologue(
            context,
            function_name=function.name,
            as_global_linker_symbol=function.is_global_linker_symbol,
        )

        amd64_linux_instruction_set(context, function.source, program, function.name)
        function_end_with_epilogue(context)


def amd64_linux_program_entry_point(context: AMD64CodegenContext) -> None:
    """Write program entry, used to not segfault due to returning into protected system memory."""
    # This is an executable entry point
    function_begin_with_prologue(
        context,
        function_name=CODEGEN_ENTRY_POINT_SYMBOL,
        as_global_linker_symbol=True,
    )

    # Prepare and execute main function
    call_function_block(
        context,
        function_name=GOFRA_ENTRY_POINT,
        abi_ffi_push_retval_onto_stack=False,
        abi_ffi_arguments_count=0,
    )

    # Call syscall to exit without accessing protected system memory.
    # `ret` into return-address will fail with segfault
    ipc_syscall_linux(
        context,
        arguments_count=1,
        store_retval_onto_stack=False,
        injected_args=[
            AMD64_LINUX_EPILOGUE_EXIT_SYSCALL_NUMBER,
            AMD64_LINUX_EPILOGUE_EXIT_CODE,
        ],
    )


def amd64_linux_data_section(
    context: AMD64CodegenContext,
    program: ProgramContext,
) -> None:
    """Write program static data section filled with static strings and memory blobs."""
    initialize_static_data_section(
        context,
        static_data_section=[
            *context.strings.items(),
            *program.memories.items(),
        ],
    )



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from gofra.parser.functions import Function

from .operators import Operator, OperatorOperand, OperatorType

if TYPE_CHECKING:
    from collections.abc import (
        Generator,
        MutableMapping,
        MutableSequence,
        Sequence,
    )
    from pathlib import Path

    from gofra.lexer import Token
    from gofra.typecheck.types import GofraType


@dataclass(frozen=False)
class ParserContext:
    """Context for parsing which only required from internal usages."""

    tokenizer: Generator[Token]

    # Should be refactored
    is_top_level: bool

    # Resulting operators from parsing
    operators: MutableSequence[Operator] = field(default_factory=lambda: list())  # noqa: C408

    functions: MutableMapping[str, Function] = field(default_factory=lambda: dict())  # noqa: C408
    memories: MutableMapping[str, int] = field(default_factory=lambda: dict())  # noqa: C408

    context_stack: deque[tuple[int, Operator]] = field(default_factory=lambda: deque())
    included_source_paths: set[Path] = field(default_factory=lambda: set())

    current_operator: int = field(default=0)

    def has_context_stack(self) -> bool:
        return len(self.context_stack) > 0

    def new_function(
        self,
        from_token: Token,
        name: str,
        *,
        type_contract_in: Sequence[GofraType],
        type_contract_out: Sequence[GofraType],
        emit_inline_body: bool,
        is_externally_defined: bool,
        is_global_linker_symbol: bool,
        source: Sequence[Operator],
    ) -> Function:
        function = Function(
            location=from_token.location,
            name=name,
            source=source,
            type_contract_in=type_contract_in,
            type_contract_out=type_contract_out,
            emit_inline_body=emit_inline_body,
            is_externally_defined=is_externally_defined,
            is_global_linker_symbol=is_global_linker_symbol,
        )
        self.functions[name] = function
        return function

    def expand_from_inline_block(self, inline_block: Function) -> None:
        if inline_block.is_externally_defined:
            msg = "Cannot expand extern function."
            raise ValueError(msg)
        self.current_operator += len(inline_block.source)
        self.operators.extend(inline_block.source)

    def pop_context_stack(self) -> tuple[int, Operator]:
        return self.context_stack.pop()

    def push_new_operator(
        self,
        type: OperatorType,  # noqa: A002
        token: Token,
        operand: OperatorOperand,
        *,
        is_contextual: bool,
    ) -> None:
        operator = Operator(
            type=type,
            token=token,
            operand=operand,
            has_optimizations=False,
        )
        self.operators.append(operator)
        if is_contextual:
            self.context_stack.append((self.current_operator, operator))
        self.current_operator += 1



from __future__ import annotations

import sys
from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path
from platform import system as current_platform_system
from typing import TYPE_CHECKING

from gofra.cli.output import cli_message

if TYPE_CHECKING:
    from gofra.assembler.assembler import OUTPUT_FORMAT_T
    from gofra.codegen.targets import TARGET_T


@dataclass(frozen=True)
class CLIArguments:
    """Arguments from argument parser provided for whole Gofra toolchain process."""

    source_filepaths: list[Path]
    output_filepath: Path
    output_format: OUTPUT_FORMAT_T

    execute_after_compilation: bool
    debug_symbols: bool

    include_paths: list[Path]

    ir: bool

    linker_flags: list[str]
    assembler_flags: list[str]

    verbose: bool

    target: TARGET_T

    disable_optimizations: bool
    skip_typecheck: bool

    build_cache_dir: Path
    delete_build_cache: bool


def parse_cli_arguments(prog: str) -> CLIArguments:
    """Parse CLI arguments from argparse into custom DTO."""

    args = _construct_argument_parser(prog=prog).parse_args()
    if len(args.source_files) > 1:
        cli_message(
            level="ERROR",
            text="Compiling several files not implemented.",
        )
        sys.exit(1)

    if None in args.include:
        cli_message(
            level="WARNING",
            text="One of the include search directories is empty, it will be skipped!",
        )
    if args.skip_typecheck:
        cli_message(
            level="WARNING",
            text="Skipping typecheck is unsafe, ensure that you know what you doing",
        )

    target: TARGET_T = args.target or infer_target()
    assert target in ("x86_64-linux", "aarch64-darwin")

    source_filepaths = [Path(f) for f in args.source_files]
    output_filepath = (
        Path(args.output)
        if args.output
        else infer_output_filename(source_filepaths, output_format=args.output_format)
    )
    include_paths = [
        Path("./"),
        *[f.parent for f in source_filepaths],
        *map(Path, [include for include in args.include if include]),
    ]

    assembler_flags = args.assembler
    if bool(args.debug_symbols):
        assembler_flags += ["-g"]

    return CLIArguments(
        debug_symbols=bool(args.debug_symbols),
        ir=bool(args.ir),
        source_filepaths=source_filepaths,
        output_filepath=output_filepath,
        output_format=args.output_format,
        execute_after_compilation=bool(args.execute),
        delete_build_cache=bool(args.delete_cache),
        build_cache_dir=Path(args.cache_dir),
        target=target,
        disable_optimizations=bool(args.disable_optimizations),
        skip_typecheck=bool(args.skip_typecheck),
        include_paths=include_paths,
        verbose=bool(args.verbose),
        linker_flags=args.linker,
        assembler_flags=assembler_flags,
    )


def _construct_argument_parser(prog: str) -> ArgumentParser:
    """Get argument parser instance to parse incoming arguments."""
    parser = ArgumentParser(
        description="Gofra Toolkit - CLI for working with Gofra programming language",
        add_help=True,
        prog=prog,
    )

    parser.add_argument(
        "source_files",
        type=str,
        help="Source code files in Gofra (`.gof` files)",
        nargs="+",
    )

    parser.add_argument(
        "--output",
        "-o",
        type=str,
        required=False,
        help="Path to output file to generate, by default will be infered from first input filename, also infers build cache filenames from that",
    )

    parser.add_argument(
        "--target",
        "-t",
        type=str,
        required=False,
        help="Compilation target. Infers codegen to use from that.",
        choices=["x86_64-linux", "aarch64-darwin"],
    )
    parser.add_argument(
        "--output-format",
        "-of",
        type=str,
        required=False,
        help="Compilation output format. Useful if you want to emit '.o' object-file.",
        default="executable",
        choices=["object", "executable", "library", "assembly"],
    )

    parser.add_argument(
        "-ir",
        required=False,
        action="store_true",
        help="If passed will just emit IR of provided file(s) into stdin.",
    )
    parser.add_argument(
        "--debug-symbols",
        "-g",
        required=False,
        action="store_true",
        help="If passed will provide debug symbols into final target output.",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        required=False,
        action="store_true",
        help="If passed will enable INFO level logs from compiler.",
    )

    parser.add_argument(
        "--execute",
        "-e",
        required=False,
        action="store_true",
        help="If provided, will execute output executable file after compilation. Expects output format to be executable",
    )

    parser.add_argument(
        "--include",
        "-i",
        required=False,
        help="Additional directories to search for include files. Default: ['./', './lib']",
        action="append",
        nargs="?",
        default=["./", "./lib"],
    )

    parser.add_argument(
        "--linker",
        "-Lf",
        required=False,
        help="Additional flags passed to linker (`ld`)",
        action="append",
        nargs="?",
        default=[],
    )

    parser.add_argument(
        "--assembler",
        "-Af",
        required=False,
        help="Additional flags passed to assembler (`as`)",
        action="append",
        nargs="?",
        default=[],
    )

    parser.add_argument(
        "--cache-dir",
        "-cd",
        type=str,
        default="./.build",
        required=False,
        help="Path to directory where to store cache (defaults to current directory)",
    )
    parser.add_argument(
        "--delete-cache",
        "-dc",
        action="store_true",
        required=False,
        help="If passed, will delete cache after run",
    )

    parser.add_argument(
        "--disable-optimizations",
        "-no",
        action="store_true",
        required=False,
        help="If passed, all optimizations will be disable (DCE, CF)",
    )

    parser.add_argument(
        "--skip-typecheck",
        "-nt",
        action="store_true",
        required=False,
        help="If passed, will disable type safety checking",
    )

    return parser


def infer_output_filename(  # noqa: PLR0911
    source_filepaths: list[Path],
    output_format: OUTPUT_FORMAT_T,
) -> Path:
    """Try to infer filename for output from input source files."""
    assert source_filepaths

    source_filepath = source_filepaths[0]

    match output_format:
        case "library":
            return source_filepath.with_suffix(".dylib")
        case "object":
            return source_filepath.with_suffix(".o")
        case "assembly":
            return source_filepath.with_suffix(".s")
        case _:
            ...
    match current_platform_system():
        case "Darwin":
            match output_format:
                case "executable":
                    if source_filepath.suffix == "":
                        return source_filepath.with_suffix("._")
                    return source_filepath.with_suffix("")
        case "Linux":
            match output_format:
                case "executable":
                    if source_filepath.suffix == "":
                        return source_filepath.with_suffix(source_filepath.suffix + "_")
                    return source_filepath.with_suffix("")
        case _:
            cli_message(
                level="ERROR",
                text="Unable to infer output filepath due to no fallback for current operating system",
            )
            sys.exit(1)


def infer_target() -> TARGET_T:
    """Try to infer target from current system."""
    assert current_platform_system() in ["Darwin", "Linux"]

    match current_platform_system():
        case "Darwin":
            return "aarch64-darwin"
        case "Linux":
            return "x86_64-linux"
        case _:
            cli_message(
                level="ERROR",
                text="Unable to infer compilation target due to no fallback for current operating system",
            )
            sys.exit(1)
